import re

from django.conf import settings
from rest_framework.throttling import SimpleRateThrottle


class ActionBasedThrottle(SimpleRateThrottle):
    __slots__ = ('scope', 'rate', 'actions')
    url_regex = re.compile(r'/(?P<scope>[a-zA-Z0-9?=&_.%]+)/([a-zA-Z0-9?=&_.%]+)?/?$')
    throttle_rates = settings.THROTTLE

    def __init__(self):  # pylint: disable=W0231 super-init-not-called
        """
        Override the usual SimpleRateThrottle, because we can't determine
        the rate until called by the view.
        """
        self.rate = self.throttle_rates['rate']
        self.actions = self.throttle_rates['actions']

    def parse_config(self, request):
        """
        Try getting per-class throttle params.If not found, getting global params.
        We take the last two parts of an URL, checking if they are present in throttle config and taking the first one,
        but in reverse order. We doing this to prevent getting id part of url, if it's presented as word.
        """
        regexed_path = self.url_regex.search(request.path)
        self.scope = next(
            (group
                for group in reversed(regexed_path.groups()) if group in self.throttle_rates['views']),
            None
        )
        if self.scope:
            self.rate = self.throttle_rates['views'][self.scope].get('rate', self.rate)
            self.actions = self.throttle_rates['views'][self.scope].get('actions', self.actions)
        else:
            self.scope = regexed_path.group(1)

    def allow_request(self, request, view):
        # do not throttle unrecognized paths
        try:
            self.parse_config(request)
        except AttributeError:
            return True
        if self.rate == '' or getattr(view, 'action', None) not in self.actions:
            return True

        self.num_requests, self.duration = self.parse_rate(self.rate)
        return super(ActionBasedThrottle, self).allow_request(request, view)

    def get_cache_key(self, request, view):
        """
        Generate the unique cache key by concatenating the user id
        with the '.throttle_scope` property of the view.
        """
        return self.cache_format % {
            'scope': self.scope,
            'ident': request.user.pk if request.user.is_authenticated else self.get_ident(request)
        }



import typing as _t

try:
    from celery import Celery
except ImportError:
    pass


CeleryType = _t.Optional[Celery]


def prepare_environment(default_settings: _t.Dict[_t.Text, _t.Text] = None, **kwargs) -> None:
    ...


def cmd_execution(*args, **kwargs) -> None:
    ...


def get_celery_app(name: _t.Text = None, **kwargs) -> CeleryType:
    ...



import os
from pathlib import Path

from ._base import BaseCommand
from ...utils import get_render
from ... import __version__


class Command(BaseCommand):
    interactive = True
    help = "Creates new project with all needed for build application."
    default_help = 'Specify the {} of project'

    _values_parser = {
        "name": {'required': True},
        "dir": {'help': 'Specify the directory where project will be.', 'default': './'},
        "guiname": {'default': ''}
    }

    files_to_create = {
        'frontend_src': {
            'app': {
                'index.js': Path('index.js')
            },
            '.editorconfig': Path('.editorconfig'),
            '.eslintrc.js': Path('.eslintrc.js'),
            '.prettierrc': Path('.prettierrc')
        },
        '{project_name}': {
            'models': {
                '__init__.py': Path('__init__.py')
            },
            '__init__.py': Path('__init__.py'),
            '__main__.py': Path('__main__.py'),
            'settings.ini': Path('settings.ini'),
            'settings.py': Path('settings.py'),
            'web.ini': Path('web.ini'),
            'wsgi.py': Path('wsgi.py'),
        },
        '.coveragerc': Path('.coveragerc'),
        '.gitignore': Path('.gitignore'),
        '.pep8': Path('.pep8'),
        'MANIFEST.in': Path('MANIFEST.in'),
        'package.json': Path('package.json'),
        'README.rst': Path('README.rst'),
        'requirements.txt': Path('requirements.txt'),
        'requirements-test.txt': Path('requirements-test.txt'),
        'setup.cfg': Path('setup.cfg'),
        'setup.py': Path('setup.py'),
        'test.py': Path('test.py'),
        'tox.ini': Path('tox.ini'),
        'tox_build.ini': Path('tox_build.ini'),
        'webpack.config.js.default': Path('webpack.config.js.default')
    }

    def add_arguments(self, parser):
        super().add_arguments(parser)
        for name, data in self._values_parser.items():
            kwargs = dict(**data)
            if not kwargs.get('help', None):
                kwargs['help'] = self.default_help.format(name)
            parser.add_argument(
                f'--{name}', dest=name, **data
            )

    def get_path(self, path, *paths) -> Path:
        return Path(os.path.expandvars(os.path.join(path, *paths))).expanduser()

    def create_file(self, render_path, path, **render_data):
        path.write_text(get_render(render_path + '.template', render_data))

    def _from_user(self, name, options, default=None):
        default = options.get(name, default)
        val_from_parser = self._values_parser.get(name, {})
        help_msg = val_from_parser.get('help', self.default_help.format(name))
        help_msg += f' Default: [{default}]'
        value = self.ask_user(help_msg, default)
        if value is None:
            raise self.CommandError(f'Value --{name} must be set.')
        return value

    def get_render_kwargs(self, options):
        project_name = self._from_user('name', options)
        project_gui_name = self._from_user('guiname', options) or project_name.upper()
        return {
            'project_name': project_name,
            'project_place': self._from_user('dir', options),
            'project_gui_name': project_gui_name,
            'vstutils_version': __version__,
            'project_gui_name_head_lines': '=' * len(project_gui_name)
        }

    def allow_create(self, path):
        root_dir_path = self.get_path(*path) if not isinstance(path, Path) else path
        if root_dir_path.exists() and len(list(root_dir_path.iterdir())) != 0:
            raise Exception(f'{root_dir_path} is not empty')

    def recursive_create(self, root, node, node_chain=None):
        if node_chain is None:
            node_chain = []

        for name, path_value in node.items():
            real_name = name.format(**self.render_vars)

            if not isinstance(path_value, Path):
                new_root = root/Path(real_name)
                new_root.mkdir()
                self.recursive_create(new_root, path_value, node_chain + [name.replace('{', '').replace('}', '')])
            else:
                template_path = str(Path('newproject', *node_chain) / path_value)
                self.create_file(template_path, root/real_name, **self.render_vars)

    def make_files(self, **render_vars):
        self.render_vars = render_vars
        self.path = self.get_path(*[render_vars['project_place'], render_vars['project_name']])
        self.allow_create(self.path)
        self.path.mkdir(exist_ok=True)
        self.recursive_create(self.path, self.files_to_create)

    def handle(self, *args, **options):
        super().handle(*args, **options)
        opts = self.get_render_kwargs(options)
        self.make_files(**opts)
        project_dir = os.path.join(opts['project_place'], opts['project_name'])
        self._print(
            f'Project successfully created at {project_dir}.', 'SUCCESS'
        )



from django.contrib.auth import get_user_model, logout
from django.contrib.auth.forms import UserCreationForm
from django.conf import settings
from django import forms
from django.forms.utils import ErrorDict
from django.test import override_settings
from django.urls import reverse
from django.core.cache import cache
from django.core.exceptions import SuspiciousOperation
from django.contrib.auth.hashers import make_password, check_password
from django.utils.html import strip_tags
from django.utils.safestring import mark_safe


from ..utils import SecurePickling, send_template_email, translate as _, lazy_translate as __

UserModel = get_user_model()
secure_pickle = SecurePickling()

override_setting_decorator = override_settings(PASSWORD_HASHERS=settings.REGISTRATION_HASHERS)
hash_data = override_setting_decorator(make_password)
check_data = override_setting_decorator(check_password)


class EmptyWidget(forms.widgets.HiddenInput):
    def render(self, *args, **kwargs):
        return ''


class AgreementWidget(forms.widgets.CheckboxInput):
    template_name = 'widgets/agreement_widget.html'

    def __init__(self, *args, **kwargs):
        kwargs['attrs'] = attrs = kwargs.get('attrs', {}) or {}
        attrs['class'] = 'form-check-input'
        super().__init__(*args, **kwargs)


class AgreementField(forms.BooleanField):
    widget = AgreementWidget

    def __init__(self,
                 before_link_text: str = '',
                 after_link_text: str = '',
                 url: str = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url_name = url
        self.before_link_text = before_link_text
        self.after_link_text = after_link_text

    def get_bootstrap_label(self, field):
        return mark_safe(  # nosec
            f'{_(self.before_link_text)}'
            f'<a href="#" onclick=open{field.name}Modal("{reverse(self.url_name)}")>'
            f'{_(self.label)}'
            f'</a>{_(self.after_link_text)}'
        )


class RegistrationForm(UserCreationForm):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if settings.ENABLE_AGREEMENT_TERMS:
            self.fields['agreement'] = AgreementField(
                before_link_text='I accept the ',
                label='terms of agreement',
                url='terms',
                error_messages={'required': __('To continue, need to accept the terms agreement.')},
            )
        if settings.ENABLE_CONSENT_TO_PROCESSING:
            self.fields['consent_to_processing'] = AgreementField(
                before_link_text='I agree with ',
                label='the personal data processing policy',
                url='terms',
                error_messages={'required': __('To continue, need to agree to the personal data processing policy.')},
            )

        for field in self.fields.values():
            if field.help_text:
                field.widget.attrs['title'] = strip_tags(field.help_text)

    error_messages = {
        'password_mismatch': __("The two password fields didn't match."),
    }

    first_name = forms.CharField(label=__('First name'), max_length=30, required=False, help_text='Optional.')
    last_name = forms.CharField(label=__('Last name'), max_length=30, required=False, help_text='Optional.')
    email = forms.EmailField(
        label=__('Email'),
        max_length=254,
        help_text=__('Required. Inform a valid email address.')  # type: ignore
    )
    if settings.SEND_CONFIRMATION_EMAIL:
        uid = forms.CharField(max_length=256, required=False, widget=EmptyWidget(), label='')
        email.help_text = __('A confirmation will be sent to your e-mail')  # type: ignore

    class Meta(UserCreationForm.Meta):
        model = UserModel
        fields = [
            'username',
            'first_name',
            'last_name',
            'email',
            'password1',
            'password2',
        ]

    def build_confirmation_url(self, uid):
        # pylint: disable=no-member
        return self.request.build_absolute_uri(f'{reverse("user_registration")}?uid={uid}')

    def get_email_context(self):
        return {}

    def register_with_confirmation(self, commit):
        cache_key = self.cleaned_data.get('uid', None)

        if cache_key in (None, ''):
            cache_key = hash_data(self.cleaned_data['email'])
            self.cleaned_data['uid'] = cache_key

            secured_data = secure_pickle.dumps(self.cleaned_data)
            cache.set(cache_key, secured_data)
            context_data = self.get_email_context()
            context_data['action_url'] = self.build_confirmation_url(cache_key)
            if context_data.get('application_name', None) is None:
                context_data['application_name'] = settings.PROJECT_GUI_NAME

            send_template_email(
                subject='Registration Confirmation.',
                template_name='registration/confirm_email.html',
                email=self.cleaned_data['email'],
                context_data=context_data
            )
            return super().save(commit=False)

        if not check_data(self.cleaned_data['email'], cache_key):
            raise SuspiciousOperation(
                _('Invalid registration email send.')
            )
        return super().save(commit)

    def save(self, commit=True):
        if settings.SEND_CONFIRMATION_EMAIL:
            return self.register_with_confirmation(commit)
        return super(RegistrationForm, self).save(commit)

    def _clean_fields(self):
        super()._clean_fields()
        if settings.SEND_CONFIRMATION_EMAIL:
            uid = self.cleaned_data.get('uid', None)
            if self.errors and uid not in (None, ''):
                try:
                    self.cleaned_data.update(secure_pickle.loads(cache.get(uid)))
                    cache.delete(uid)
                    self._errors = ErrorDict()
                except TypeError:  # nocv
                    self._errors = ErrorDict()
                    self.add_error('uid', _('Confirmation link is invalid or expired'))

    def _post_clean(self):
        super()._post_clean()
        if settings.SEND_CONFIRMATION_EMAIL and 'uid' in self._errors:
            for err in tuple(self._errors.keys()):
                if err != 'uid':
                    del self._errors[err]


class TwoFaForm(forms.Form):
    pin = forms.CharField(required=True)

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop('request', None)
        super(TwoFaForm, self).__init__(*args, **kwargs)

    def clean(self):
        pin = self.cleaned_data.get('pin')

        if self.request.user.twofa.verify(pin):
            self.request.session['2fa'] = True
        else:
            self.add_error('pin', 'Invalid authentication code')
            attempts = self.request.session.get('2fa_attempts', 0) + 1
            if attempts >= settings.MAX_TFA_ATTEMPTS:
                logout(self.request)
            else:
                self.request.session['2fa_attempts'] = attempts

        return super().clean()

    def get_user(self):
        return self.request.user



# pylint: disable=no-member,unused-argument
"""
Default serializer classes for web-api.
Read more in Django REST Framework documentation for
`Serializers <https://www.django-rest-framework.org/api-guide/serializers/>`_.
"""


import json

from django.db import models
from django.http.request import QueryDict
from rest_framework import serializers
from rest_framework.utils.field_mapping import get_relation_kwargs

from . import fields
from .. import utils
from ..models.fields import (
    NamedBinaryFileInJSONField,
    NamedBinaryImageInJSONField,
    MultipleNamedBinaryFileInJSONField,
    MultipleNamedBinaryImageInJSONField,
    MultipleFileField,
    MultipleImageField,
    FkModelField
)

VALID_FK_KWARGS = (
    'read_only',
    'label',
    'help_text',
    'allow_null',
    'required'
)


class BaseSerializer(serializers.Serializer):
    """
    Default serializer with logic to work with objects.
    Read more in `DRF serializer's documentation
    <https://www.django-rest-framework.org/api-guide/serializers/#serializers>`_
    how to create Serializers and work with them.
    """

    # pylint: disable=abstract-method

    def create(self, validated_data):  # nocv
        return validated_data

    def update(self, instance, validated_data):  # nocv
        if isinstance(instance, dict):
            instance.update(validated_data)
        else:
            for key, value in validated_data.items():
                setattr(instance, key, value)
        return instance


class VSTSerializer(serializers.ModelSerializer):
    """
    Default model serializer based on :class:`rest_framework.serializers.ModelSerializer`.
    Read more in `DRF documentation <https://www.django-rest-framework.org/api-guide/serializers/#modelserializer>`_
    how to create Model Serializers.
    This serializer matches model fields to extended set of serializer fields.
    List of available pairs specified in  `VSTSerializer.serializer_field_mapping`.
    For example, to set :class:`vstutils.api.fields.FkModelField` in serializer use
    :class:`vstutils.models.fields.FkModelField` in a model.
    """
    # pylint: disable=abstract-method

    serializer_field_mapping = serializers.ModelSerializer.serializer_field_mapping
    serializer_field_mapping.update({
        models.CharField: fields.VSTCharField,  # type: ignore
        models.TextField: fields.VSTCharField,  # type: ignore
        models.FileField: fields.NamedBinaryFileInJsonField,  # type: ignore
        models.ImageField: fields.NamedBinaryImageInJsonField,  # type: ignore
        NamedBinaryFileInJSONField: fields.NamedBinaryFileInJsonField,  # type: ignore
        NamedBinaryImageInJSONField: fields.NamedBinaryImageInJsonField,  # type: ignore
        MultipleNamedBinaryFileInJSONField: fields.MultipleNamedBinaryFileInJsonField,  # type: ignore
        MultipleNamedBinaryImageInJSONField: fields.MultipleNamedBinaryImageInJsonField,  # type: ignore
        MultipleFileField: fields.MultipleNamedBinaryFileInJsonField,  # type: ignore
        MultipleImageField: fields.MultipleNamedBinaryImageInJsonField  # type: ignore
    })

    def build_standard_field(self, field_name, model_field):
        field_class, field_kwargs = super().build_standard_field(field_name, model_field)
        if isinstance(model_field, models.FileField) and issubclass(field_class, fields.NamedBinaryFileInJsonField):
            field_kwargs['file'] = True
        return field_class, field_kwargs

    def build_relational_field(self, field_name, relation_info):
        if isinstance(relation_info.model_field, FkModelField) and \
                hasattr(relation_info.related_model, '__extra_metadata__'):

            field_kwargs = {
                key: value
                for key, value in get_relation_kwargs(field_name, relation_info).items()
                if key in VALID_FK_KWARGS
            }
            field_kwargs['select'] = relation_info.related_model

            return fields.FkModelField, field_kwargs
        # if DRF ForeignField in model or related_model is not BModel, perform default DRF logic
        return super().build_relational_field(field_name, relation_info)


class EmptySerializer(BaseSerializer):
    """
    Default serializer for empty responses.
    In generated GUI this means that action button which will not show additional view before execution.
    """


class DataSerializer(EmptySerializer):
    allowed_data_types = (
        str,
        dict,
        list,
        tuple,
        type(None),
        int,
        float
    )

    def to_internal_value(self, data):
        if isinstance(data, QueryDict):
            return data.dict()  # nocv
        return data if isinstance(data, self.allowed_data_types) else self.fail("Unknown type.")

    def to_representation(self, instance):
        if not isinstance(instance, (dict, list)):
            result = json.loads(instance)
            if isinstance(result, dict):
                result = utils.Dict(result)
            return result
        return instance


class JsonObjectSerializer(DataSerializer):
    pass


class ErrorSerializer(DataSerializer):
    detail = fields.VSTCharField(required=True)

    def to_internal_value(self, data):
        return data

    def to_representation(self, instance):
        return instance


class ValidationErrorSerializer(ErrorSerializer):
    detail = serializers.DictField(required=True)  # type: ignore


class OtherErrorsSerializer(ErrorSerializer):
    error_type = fields.VSTCharField(required=False, allow_null=True)



# Generated by Django 2.2.18 on 2021-03-02 08:27

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0011_update_proxy_permissions'),
        ('vstutils_api', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='TwoFactor',
            fields=[
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='twofa', serialize=False, to=settings.AUTH_USER_MODEL)),
                ('secret', models.CharField(max_length=256)),
            ],
            options={
                'default_related_name': 'twofa',
            },
        ),
        migrations.CreateModel(
            name='RecoveryCode',
            fields=[
                ('id', models.AutoField(max_length=100, primary_key=True, serialize=False)),
                ('key', models.CharField(max_length=256)),
                ('tfa', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='recoverycode', to='vstutils_api.TwoFactor')),
            ],
            options={
                'default_related_name': 'recoverycode',
            },
        ),
    ]



# pylint: disable=unused-import
from rest_framework import status


class VSTUtilsException(Exception):
    msg = "Bad request."
    status: int = status.HTTP_400_BAD_REQUEST

    def __init__(self, *args, **kwargs):
        self.msg = (list(args)[0:1]+[""])[0]
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return str(self.msg)


class UnknownTypeException(VSTUtilsException):
    _def_message = "Unknown type {}."
    status = status.HTTP_415_UNSUPPORTED_MEDIA_TYPE

    def __init__(self, tp, msg=None):
        self._def_message = msg or self._def_message
        msg = self._def_message.format(tp)
        super().__init__(msg)


class NotApplicable(VSTUtilsException):
    status = status.HTTP_404_NOT_FOUND



# flake8: noqa: E501
TRANSLATION = {
    "new": '新的',
    "Create": '建立',
    "add": '增加',
    "Edit": '校对',
    "edit": '校对',
    "Save": '储存',
    "Remove": '删除',
    "Reload": '重启',
    "send": '发送',
    "select": '选择',
    "filters": '过滤',
    "apply": '应用',
    "cancel": '取消',
    "field": '范围',
    "actions": '活动',
    "sublinks": '输入链接',
    "name": '种类',
    "type": '类型',
    "owner": '业主',
    "notes": '标记',
    "copy": '复制',
    "error": '错误',
    "home": '主要的',
    "login": '进入',
    "logout": '退出',
    "Profile": '轮廓',
    "settings": '设置',
    "Settings": '设置',
    "version": '版本',
    "cache": '隐藏',
    "documentation": '文件',
    "request": '询问',
    "repository": '复位器',
    "app info": '信息',
    "status": '状态',
    "List is empty": '空白表格',
    "child instances": '输入对象',
    "enter value": '输入意义',
    "more info": '输入价值',
    "search by": '更详尽',
    'Search': '搜索 ',
    'Selected': '选中',
    'Nothing found!': '没有发现！',
    'Main settings': '主要设置',
    'Custom settings': '自定义设置',
    'Remove from list': '從列表中刪除',
    'Purge': '刪除',
    'Select an action for this instances': '為此實例選擇一個操作',
    'Username not': '用户名不是',


    # fields
    'Id': '标识符',
    'ID': '标识符',

    # x-menu generic
    'system': '范围查找',
    'users': '系统',
    'Is active': '活跃度',
    'Is staff': '减速装置',

    # multiactions button
    'actions on': '行动结束',
    'on item': '元素| 元素| 元素| 元素',

    'key': '要素',
    'value': '意义',
    'new owner': '新业主',
    'detail': '细节',
    'description': '描述',
    'username': '使用者',
    'is active': '活跃度',
    'user': '使用者',
    'yes': '是',
    'no': '否',
    'is staff': '管理',
    'first name': '名字',
    'last name': '姓',
    'password': '密码',
    'Password': '密码',
    'Change password': '变换密码',
    'Change_password': '变换密码',
    'generate password': '形成密码',
    'repeat password': '重复密码',
    'Repeat password': '重复密码',
    'Old password': '旧密码',
    'New password': '新密码',
    'Confirm new password': '确认密码',
    'kind': '类型',
    'mode': '制度',
    'options': '选项',
    'information': '信息',
    'date': '日期',

    'file n selected': '未选择文件| 已选择1个文件| 已选择{n}个文件| 已选择{n}个文件',
    'image n selected': '未选择图像 已选择1张图片| 已选择{n}张图片| 已选择{n}张图片',

    # filters
    'Ordering': '排列顺序',
    'is_active': '活跃度',
    'Username': '使用名',
    'First name': '名字',
    'Last name': '姓',

    'Page matching current url was not found': '找不到与指定网址匹配的页面',
    'Unknown error': '未知错误',
    'Error': '错误',

    'We\'ve emailed you instructions for setting your password, if an account exists with the email you entered. You should receive them shortly.': '我们已向您发送了一封电子邮件，其中包含有关重设密码的说明',
    'If you don\'t receive an email, please make sure you\'ve entered the address you registered with, and check your spam folder.': '如果您尚未收到电子邮件，请确认输入正确的地址并检查垃圾邮件文',
    'The password reset link was invalid, possibly because it has already been used. Please request a new Password reset.': '无效的密码恢复链接，可能已被使用。 请再试一次。',
    'Please enter your new password twice so we can verify you typed it in correctly.': '请输入新密码',
    'change my password': '换密码',
    'Your password has been set.  You may go ahead and log in now.': '您的密码已成功设置。 您现在可以登录',
    'Please enter a correct %(username)s and password. Note that both fields may be case-sensitive.': '请输入正确的％（用户名）和密码。 注意，这些字段可能区分大小写',
    'or': '或者',
    'two-factor authentication code': '两因素授权码',
    'verify code': '确认密码',
    'Enter the code from the two-factor app on your mobile device. If you\'ve lost your device, you may enter one of your recovery codes.': '从您的移动设备上的两个应用程序中输入代码。 如果您丢失了设备，则可以输入恢复其中一个代码',
    'Scan the image with the two-factor authentication app on your phone. If you can’t use a barcode, enter this text code instead.': '使用手机上的两个身份验证应用程序扫描图像。 如果您不能使用条形码，请输入文本代码。',
    'Enter the six-digit code from the application': '从应用程序中输入六个字符的代码',
    'code': '代码',
    'recovery codes': '恢复代码',
    'Recovery codes are used to access your account in the event you cannot receive two-factor authentication codes.': '如果您无法获得两个身份验证码，则可以使用恢复码来访问您的帐户。',
    'Disabling Two-Factor Authentication (2FA) will decrease your account\'s security': '禁用两个身份验证会降低您帐户的安全性',
    'Enable': '打开',
    'Disable': '关闭',
    'Two factor authentication': '两个验证',
    'No {0} matches the given query.': '{0}没有一个与此请求匹配',
    'Password is not correct.': '密码不正确',
    'New passwords values are not equal.': '新密码值不相同',
    'This password is entirely numeric.': '该密码应使用完全数字',
    'Passwords do not match.': '密码不匹配。',
    "The two password fields didn't match.": '两个密码字段不匹配。',
    'Dark theme': '黑色主题',
    'Language': '语言',
    'height': '高度',
    'width': '宽度',
    'More info': '更多信息',
    'Dark mode': '黑暗主题',
    'Execute': '施行',
    'Domain should be setuped or username should be with @domain.name': "应设置域或用户名应为@domain.name",
    'Invalid name': '无效名称',
    'Should be [full.domain.name]': '应该是 [full.domain.name].',
    'Invalid registration email send.': '无效的注册电子邮件发送。',
    'To continue, need to accept the terms agreement.': '要继续，需要接受条款协议。',


    # Проверить
    'Changes in settings are successfully saved. Please refresh the page.': '设置更改已成功保存。 请刷新页面。',
    'now': '眼下',
    'later': '之后',
    'I accept the ': '我接受 ',
    'Zoom': '飛漲',
    'Scale': '規模',
    'This file format is not supported': '不支持此文件格式',
    'Oops! Something went wrong.': '哎呀！ 出了些问题。',
    '[Object not found]': '[未找到对象]',
    'Execute action on {0}': '在 {0} 上执行操作',
    '{n} instance': '{n} 个实例',
    'Open image': '打开图片',
    'Confirm action': '确认操作',
    'Confirm': '确认',
    'Cancel': '取消',

    # field's validation
    # empty
    'Field "<b>{0}</b>" is empty.': '范围"<b>{0}</b>" 是空的',
    # required
    'Field "<b>{0}</b>" is required.': '范围"<b>{0}</b>" 必须完成',
    # minLength
    'Field "<b>{0}</b>" is too short.<br> Field length should not be shorter, then {1}.': '长度范围 "<b>{0}</b>" 太少.<br>长度范围应不少于{1}.',
    # maxLength
    'Field "<b>{0}</b>" is too long. <br> Field length should not be longer, then {1}.': '长度单位"<b>{0}</b>" 太大<br>长度单位 应不大于{1}.',
    # min
    'Field "<b>{0}</b>" is too small.<br> Field should not be smaller, then {1}.': '取值范围 "<b>{0}</b>" 太少.<br> 取值范围应不少于 {1}.',
    # max
    'Field "<b>{0}</b>" is too big.<br> Field should not be bigger, then {1}.': '取值范围 "<b>{0}</b>" 太大.<br> 取值范围应不大于{1}.',
    # invalid
    '<b>{0} </b> value is not valid for <b>{1}</b> field.': '不能在 <b>{1}</b>取值为<b>{0}</b> .',
    # email
    '<b>"{0}"</b> field should be written in <b>"example@mail.com"</b> format.': '应在 <b>"example@mail.com"</b>.范围内完成<b>"{0}"</b>',

    # validators
    'imageValidationResolutionError': "必须至少为{min}且不超过{max}个像素",

    # instance operation success
    # add
    'Child "<b>{0}</b>" instance was successfully added to parent list.': '子对象“ <b> {0} </ b>”已成功添加到母列表中.',
    # create
    'New "<b>{0}</b>" instance was successfully created.': '新对象“ <b> {0} </ b>”已成功创建.',
    # remove
    '"<b>{0}</b>" {1} was successfully removed.': '"<b>{0}</b>" {1} 已成功删除."',
    # save
    'Changes in "<b>{0}</b>" {1} were successfully saved.': '对象变化{1}“ <b> {0} </ b>”的更改已成功保存.',
    # execute
    'Action "<b>{0}</b>" was successfully executed on "<b>{1}</b>" instance.': '在对象“ <b> {1} </ b>”上成功启动了操作“ <b> {0} </ b>”。',
    # instance operation error
    # add:
    'An error occurred during adding of child "<b>{0}</b>" instance to parent list.<br> Error details: {1}': '将对象“ <b> {0} </ b>”添加到目级列表时发生错误。<br>更多详细信息：{1}',
    # create:
    'An error occurred during creation.<br>Error details:<br>{0}': '创建新对象时发生错误<br>更多详细信息：{0}',
    # remove:
    'An error occurred during removal process of "<b>{0}</b>" {1}.<br> Error details: {2}': '删除{1}“ <b> {0} </ b>”时发生错误<br>更多详细信息：{2}',
    # save:
    'An error occurred during process.<br>Error details:<br>{0}': '保存时发生错误。<br>更多详细信息：{0}',
    # execute:
    'An error occurred during <b>{0}</b>. Error details:<br>{1}': '<b>{0}</b> 期间发生了一些错误。 错误详情：<br>{1}',
    'Link': '链接',

    # csv
    'Actions': '行動',
    'Index': '指數',
    'Download with headers': '帶標題下載',
}

SERVER_TRANSLATION = {
    # field's validation
    "value is not a valid list": "值不是有效列表",
    "value is not a valid JSON": "值不是有效的 JSON",
    "key {missing_key} is missing": "缺少密钥 {missing_key}",
    "invalid key {invalid_key}": "无效密钥 {invalid_key}",

    # field help_text
    'A confirmation will be sent to your e-mail': '确认将发送到您的电子邮箱',
    'Required. Inform a valid email address': '必需的。 通知一个有效的电子邮件地址',

    # validators
    'unsupported image file format, expected ({0}), got {1}': '不支持的图像格式，支持（{0}），已收到{1}',
    'For some reason, this image file cannot be opened': '无法打开图像文件',
    'Invalid image {0}. Expected from {1} to {2}, got {3}': "{0}图片不正确。 预期从{1}到{2}，收到{3}",
    'This field must contain only digits but input: ': '此字段必须仅包含数字但输入：',
    'Unsupported image file format': '不支持的图像文件格式',
    'is not in listed supported types': '不在列出的受支持类型中',
    'Invalid image size orientations': '图片尺寸不正确',
    'Current image size': '当前图像大小',

    # agreements
    'personal data processing policy': '个人资料处理政策',
    'I agree with ': '我同意 ',
    'the personal data processing policy': '个人资料处理政策',
    'To continue, need to agree to the personal data processing policy.': '要继续，需要同意个人数据处理政策。',
    'terms of agreement': '协议条款',

    #auth/reg
    'sign in to start your session': '登录开始会议',
    'confirm your email before logging in': '登录前确认您的电子邮件',
    'Confirmation link is invalid or expired': '确认链接无效或过期',
    'sign in': '进入',
    'sign up': '请注册',
    'forgot password': '忘记密码',
    'confirm password': '确认密码',
    'register': '注册',
    'fields with * is required': '标有*为必填项',
    'Forgot your password? Enter your email address below, and an email with instructions for setting a new one will be sent.': '忘记密码了吗？ 输入您的地址，我们将向您发送恢复指示',
    'reset my Password': '问密码',
    'Invalid authentication code': '无效的授权码',
    'Secret string must be provided': '必须提供秘密字符串',

    # filters help text
    'A unique integer value (or comma separated list) identifying this instance.': '标识此对象的唯一数值（或用逗号分隔的序列）识别此对象',
    'A name string value (or comma separated list) of instance.': '对象名称-字符串值（或它们的顺序，用逗号分隔）',
    'Which field to use when ordering the results.': '结果排序依据的字段',
    "Search by {}'s primary key or {}": '按 {} 的主键或 {} 搜索',
    '_name': '种类',
    'Boolean value meaning status of user.': '指示用户状态的布尔值。',
    'Users first name.': '用户名',
    'Users last name.': '用户姓',
    'Users e-mail value.': '用户邮箱',
}



# pylint: disable=protected-access
import typing as _t
from collections import OrderedDict
from inspect import getmembers
import json

from django.db import transaction, models
from rest_framework.decorators import action
from rest_framework import response, request as drf_request, status, views, serializers
from drf_yasg.utils import swagger_auto_schema

from . import base
from .filter_backends import DeepViewFilterBackend
from ..exceptions import VSTUtilsException
from .. import utils

MasterViewType = _t.Type[base.GenericViewSet]


def ensure_is_object(obj):
    if isinstance(obj, str):
        return json.loads(obj)
    return obj


def _is_extra_action(attr):
    return hasattr(attr, 'mapping')


def _create_with_iteration(view, get_serializer_func, nested_manager, data):
    serializer = get_serializer_func(data=data)
    serializer.is_valid(raise_exception=True)

    if hasattr(nested_manager, 'field'):
        # on fk relations
        serializer.validated_data[nested_manager.field.name] = nested_manager.instance
        view.perform_create(serializer)
        return serializer.instance

    elif hasattr(nested_manager, 'content_type_field_name'):
        # on contenttype relations
        content_type_field_name = nested_manager.content_type_field_name
        serializer.validated_data[content_type_field_name] = getattr(nested_manager, content_type_field_name)
        serializer.validated_data[nested_manager.object_id_field_name] = nested_manager.instance.id
        view.perform_create(serializer)
        return serializer.instance

    # other relations (many to many or custom with `.add()` method)
    with transaction.atomic():
        view.perform_create(serializer)
        nested_manager.add(serializer.instance)
    return serializer.instance


def __get_nested_path(
        name: _t.Text,
        arg: _t.Text = None,
        arg_regexp: _t.Text = '[0-9]',
        empty_arg: bool = True) -> _t.Text:
    path = name
    if not arg:
        return path
    path = ''.join([
        path, '/?(?P<', arg, '>', arg_regexp, '*' if empty_arg else "+", ')'
    ])
    return path


def __get_nested_subpath(*args, **kwargs) -> _t.Text:
    sub_path = kwargs.pop('sub_path', None)
    path = __get_nested_path(*args, **kwargs)
    if sub_path:
        path += '/'
        path += sub_path
    return path


def nested_action(
        name: _t.Text,
        arg: _t.Text = None,
        methods=None,
        manager_name=None,
        *args, **kwargs) -> _t.Callable:
    # pylint: disable=too-many-locals
    list_methods: _t.List[_t.Text] = ['get', 'head', 'options', 'post']
    detail_methods: _t.List[_t.Text] = ['get', 'head', 'options', 'put', 'patch', 'delete']
    methods = list(methods or (detail_methods if arg else list_methods))
    arg_regexp: _t.Text = kwargs.pop('arg_regexp', '[0-9]')
    empty_arg: bool = kwargs.pop('empty_arg', True)
    request_arg = kwargs.pop('request_arg', f'{name}_{arg}')
    request_arg = request_arg if arg else None
    append_arg = kwargs.pop('append_arg', arg)
    sub_options = kwargs.pop('sub_opts', {})
    path = __get_nested_subpath(name, request_arg, arg_regexp, empty_arg, **sub_options)
    allow_append = bool(kwargs.pop('allow_append', False))
    manager_name = manager_name or name
    _nested_args = kwargs.pop('_nested_args', OrderedDict())
    _nested_filter_class = kwargs.pop('filterset_class', None)

    def decorator(func: _t.Callable):
        # pylint: disable=used-before-assignment
        def wrapper(view: NestedViewMixin, request: drf_request.Request, *args, **kwargs):
            # Nested name
            view.nested_name = name
            # Allow append to nested or only create
            view.nested_allow_append = allow_append
            # ID name of nested object
            view.nested_arg = request_arg
            view.nested_append_arg = append_arg
            view.nested_id = kwargs.get(view.nested_arg, None)
            view.nested_view_object = None
            view._nested_filter_class = _nested_filter_class
            return func(view, request, *args)

        wrapper.__name__ = func.__name__
        kwargs['methods'] = methods
        kwargs['detail'] = True
        kwargs['url_path'] = path
        kwargs['url_name'] = kwargs.pop('url_name', name)
        view: NestedViewMixin = action(*args, **kwargs)(wrapper)  # type: ignore
        view._nested_args = _nested_args
        view._nested_manager = manager_name or name
        view._nested_filter_class = _nested_filter_class
        if arg:
            view._nested_args[name] = request_arg
        return view

    return decorator


def subaction(*args, **kwargs):
    """
    Decorator which wrap object method to subaction of viewset.

    :param methods: List of allowed HTTP-request methods. Default is ``["post"]``.
    :type methods: list
    :param detail: Flag to set method execution to one instance.
    :type detail: bool
    :param serializer_class: Serializer for this action.
    :type serializer_class: vstutils.api.serializers.VSTSerializer
    :param permission_classes: Tuple or list permission classes.
    :param url_path: API-path name for this action.
    :type url_path: str
    :param description: Description for this action in OpenAPI.
    :type description: str
    :param multiaction: Allow to use this action in multiactions.
                        Works only with :class:`vstutils.api.serializers.EmptySerializer` as response.
    :type multiaction: bool
    :param require_confirmation: Sets whether the action must be confirmed before being executed.
    :type require_confirmation: bool
    """

    operation_description = kwargs.pop('description', None)
    response_code = kwargs.pop('response_code', None)
    serializer_class = kwargs.get('serializer_class', None)
    response_serializer = kwargs.pop('response_serializer', serializer_class)

    assert (
        (response_code is None) or
        (response_code is not None and response_serializer is not None)
    ), "If `response_code` was setted, `response_serializer` should be setted too."

    is_mul = kwargs.pop('multiaction', False)
    require_confirmation = kwargs.pop('require_confirmation', False)
    kwargs['methods'] = kwargs.pop('methods', ['post'])

    def decorator(func: _t.Callable):
        func_object = action(*args, **kwargs)(func)
        override_kw: _t.Dict = {'methods': tuple(func_object.mapping.keys()) or None}  # type: ignore

        if response_code:
            override_kw['responses'] = {
                response_code: response_serializer()
            }

        if operation_description:
            override_kw['operation_description'] = operation_description
        else:
            override_kw['operation_description'] = str(func.__doc__ or '').strip()  # type: ignore

        override_kw['x-multiaction'] = bool(is_mul)

        override_kw['x-require-confirmation'] = bool(require_confirmation)

        return swagger_auto_schema(**override_kw)(func_object)  # type: ignore

    return decorator


def get_action_name(master_view: MasterViewType, method: _t.Text = '') -> _t.Text:
    method = method.lower()
    if method == 'post':
        action_name = 'create'
    elif method == 'get' and not master_view.nested_detail:
        action_name = 'list'
    elif method == 'get' and master_view.nested_detail:
        action_name = 'retrieve'
    elif method == 'put':
        action_name = 'update'
    elif method == 'patch':
        action_name = 'partial_update'
    elif method == 'delete':
        action_name = 'destroy'
    else:  # nocv
        action_name = None  # type: ignore

    return action_name


class NestedViewMixin:
    __slots__ = ('action', 'headers')
    get_serializer: _t.Callable
    check_object_permissions: _t.Callable
    check_permissions: _t.Callable
    check_throttles: _t.Callable
    lookup_field: _t.Text
    request: drf_request.Request
    nested_name: _t.Text
    nested_append_arg: _t.Text
    nested_allow_append: bool
    nested_arg: _t.Text
    nested_id: _t.Union[_t.Text, int]
    nested_view_object: _t.Optional[models.Model]
    nested_parent_object: _t.Optional[models.Model]
    nested_detail: bool
    _nested_filter_class: _t.Any
    _nested_args: _t.Dict[_t.Text, _t.Any]
    _nested_manager: _t.Union[models.QuerySet, _t.Text]
    nested_manager: _t.Union[models.Manager, models.QuerySet]
    queryset_filters: _t.List
    master_view: MasterViewType
    kwargs: _t.Dict
    get_view_methods: _t.ClassVar[_t.Callable]
    headers: _t.Dict
    default_response_headers: _t.Dict
    finalize_response: _t.ClassVar[_t.Callable]
    check_etag: _t.ClassVar[_t.Callable]
    is_main_action: bool

    def _check_permission_obj(self, objects: _t.Iterable):
        for obj in objects:
            self.check_object_permissions(self.request, obj)

    def get_queryset(self) -> models.QuerySet:
        qs = self.nested_manager.all()
        if any(
            DeepViewFilterBackend in getattr(self, attr, [])
            for attr in ['pre_filter_backends', 'filter_backends']
        ):
            qs = DeepViewFilterBackend().filter_queryset(self.request, qs, self)
        for qs_filter in self.queryset_filters:
            if callable(qs_filter):
                qs = qs_filter(self.nested_parent_object, qs)
        return qs

    def get_nested_action_name(self) -> _t.Text:
        return get_action_name(self.master_view, str(self.request.method))

    def get_serializer_context(self) -> _t.Dict:
        context = super().get_serializer_context()  # type: ignore
        return context

    def perform_destroy(self, instance):
        purge_nested = self.master_view.request.headers.get('X-Purge-Nested', 'false') == 'true'

        if self.master_view.nested_allow_append and not purge_nested:  # type: ignore
            # pylint: disable=import-outside-toplevel
            from vstutils.models import notify_clients

            self.nested_manager.remove(instance)
            notify_clients(instance.__class__, instance.pk)
            notify_clients(self.nested_parent_object.__class__, self.nested_parent_object.pk)
        else:
            instance.delete()

    @transaction.atomic()
    def dispatch_route(self, nested_sub=None) -> base.RestResponse:
        kwargs = {}
        if nested_sub:
            self.action = nested_sub
        else:
            self.action = self.get_nested_action_name()
        if self.action != 'list':
            kwargs[self.nested_append_arg] = self.nested_id
        self.check_permissions(self.request)
        self.check_throttles(self.request)
        if getattr(self, 'is_main_action', False):
            getattr(self, 'check_etag', lambda r: None)(self.request)
        return self.finalize_response(
            self.request,
            getattr(self, self.action)(self.request, **kwargs),
            **kwargs
        )


class NestedWithoutAppendMixin(NestedViewMixin):
    __slots__ = ()

    def create(self, request: drf_request.Request, *args, **kwargs):
        # pylint: disable=unused-argument
        many = isinstance(request.data, (list, tuple))
        return self.perform_create_nested(request.data, self.lookup_field, many)

    def prepare_request_data(self, request_data, many):
        return request_data if many else [request_data]

    def _data_create(self, request_data, nested_append_arg):
        get_serializer_func, manager = self.get_serializer, self.nested_manager
        return [
            getattr(_create_with_iteration(self, get_serializer_func, manager, d), nested_append_arg)
            for d in request_data
        ]

    def perform_create_nested(self, request_data, nested_append_arg, many) -> base.RestResponse:
        id_list = self._data_create(
            self.prepare_request_data(request_data, many), nested_append_arg
        )
        qs_filter = {nested_append_arg + '__in': id_list}
        queryset = self.get_queryset().filter(**qs_filter)

        if not many:
            queryset = queryset.get()

        serializer = self.get_serializer(queryset, many=many)
        return response.Response(serializer.data, status=status.HTTP_201_CREATED)


class NestedWithAppendMixin(NestedWithoutAppendMixin):
    __slots__ = ()

    def _data_create(self, request_data, nested_append_arg):
        # pylint: disable=import-outside-toplevel
        from vstutils.models import bulk_notify_clients

        filter_arg = f'{nested_append_arg}__in'
        request_data = [ensure_is_object(d) for d in request_data]
        objects = self.get_queryset().model.objects.filter(**{
            filter_arg: map(lambda i: i.get(nested_append_arg), request_data)
        })
        self._check_permission_obj(objects)
        self.nested_manager.add(*objects)
        id_list = [o.pk for o in objects]

        label = objects.model._meta.label

        handler = self.get_serializer_class()().get_fields()[nested_append_arg].to_representation

        def is_not_created(data):
            try:
                return handler(data.get(nested_append_arg, None)) not in id_list
            except:
                return True

        notif_objects = tuple(
            (label, pk)
            for pk in id_list
        )
        id_list += super()._data_create(
            filter(is_not_created, request_data), nested_append_arg
        )

        bulk_notify_clients(objects=notif_objects)
        return id_list


def nested_view_function(
        master_view: _t.Union[base.GenericViewSet, NestedViewMixin],
        view: _t.Type[_t.Union[NestedViewMixin, base.GenericViewSet, views.APIView]],
        view_request: drf_request.Request,
        *args, **kw) -> base.RestResponse:
    # pylint: disable=unused-argument,unnecessary-lambda
    nested_sub = kw.get('nested_sub', None)
    view_obj: _t.Union[NestedViewMixin, base.GenericViewSet, views.APIView] = view(  # type: ignore
        request=view_request,
        kwargs=view_request.parser_context['kwargs'],  # type: ignore
        headers=master_view.headers,
    )
    master_view.nested_view_object = view_obj  # type: ignore
    master_view.nested_detail = view_obj.nested_detail  # type: ignore
    return view_obj.dispatch_route(nested_sub)  # type: ignore


class BaseClassDecorator:
    __slots__ = 'name', 'arg', 'request_arg', 'args', 'kwargs'
    name: _t.Text
    request_arg: _t.Text
    args: _t.Tuple
    kwargs: _t.Dict[str, _t.Any]

    def __init__(self, name: _t.Text, arg: _t.Text, *args, **kwargs):
        self.name = name
        self.arg = arg
        self.request_arg = kwargs.pop('request_arg', f'{self.name}_{self.arg}')
        self.args = args
        self.kwargs = kwargs

    def setup(self, view_class):  # nocv
        raise NotImplementedError()

    def __call__(self, view_class):
        return self.decorator(view_class)

    def decorator(self, view_class):
        return self.setup(view_class)


class nested_view(BaseClassDecorator):  # pylint: disable=invalid-name
    """
    By default DRF does not support nested views.
    This decorator solves this problem.

    You need two or more models with nested relationship (Many-to-Many or Many-to-One)
    and two viewsets. Decorator nests viewset to parent viewset class and
    generate paths in API.

    :param name: -- Name of nested path. Also used as default name for related queryset (see `manager_name`).
    :type name: str
    :param arg: -- Name of nested primary key field.
    :type arg: str
    :param view: -- Nested viewset class.
    :type view:
        vstutils.api.base.ModelViewSet,
        vstutils.api.base.HistoryModelViewSet,
        vstutils.api.base.ReadOnlyModelViewSet
    :param allow_append: -- Flag for allowing to append existed instances.
    :type allow_append: bool
    :param manager_name: -- Name of model-object attr which contains nested queryset.
    :type manager_name: str
    :param methods: -- List of allowed methods to nested view endpoints.
    :type methods: list
    :param subs: -- List of allowed subviews or actions to nested view endpoints.
    :type subs: list,None


    .. note::
        Some view methods will not calling for performance reasons.
        This also applies to some of the class attributes that are usually initialized in the methods.
        For example, ``.initial()`` will never called.
        Each viewset wrapped by nested class with additional logic.

    Example:

        .. sourcecode:: python

            from vstutils.api.decorators import nested_view
            from vstutils.api.base import ModelViewSet
            from . import serializers as sers


            class StageViewSet(ModelViewSet):
                model = sers.models.Stage
                serializer_class = sers.StageSerializer


            nested_view('stages', 'id', view=StageViewSet)
            class TaskViewSet(ModelViewSet):
                model = sers.models.Task
                serializer_class = sers.TaskSerializer


        This code generates api paths:

         * `/tasks/` - GET,POST
         * `/tasks/{id}/` - GET,PUT,PATCH,DELETE
         * `/tasks/{id}/stages/` - GET,POST
         * `/tasks/{id}/stages/{stages_id}/` - GET,PUT,PATCH,DELETE

    """
    __slots__ = (
        'view',
        'allowed_subs',
        '_subs',
        'serializers',
        'methods',
        'queryset_filters'
    )
    view: _t.Optional[_t.Type[_t.Union[NestedViewMixin, base.GenericViewSet]]]
    allowed_subs: _t.Optional[_t.Union[_t.List[_t.Text], _t.Tuple]]
    serializers: _t.Tuple[serializers.Serializer, serializers.Serializer]
    methods: _t.Optional[_t.List[_t.Text]]
    queryset_filters: _t.List

    filter_subs = ['filter']

    class NoView(VSTUtilsException):
        msg = 'Argument "view" must be installed for `nested_view` decorator.'

    def __init__(self, name, arg=None, methods=None, *args, **kwargs):
        self.view = kwargs.pop('view', None)
        self.allowed_subs = kwargs.pop('subs', ())
        self.queryset_filters = kwargs.pop('queryset_filters', [])
        super().__init__(name, arg, *args, **kwargs)
        self._subs = set(self.get_subs())

        if self.view is None:
            raise self.NoView()

        self.serializers = self.__get_serializers(kwargs)
        self.methods = methods

        if self.arg is None:
            self.methods = methods or ['get']

        self.kwargs['empty_arg'] = self.kwargs.pop('empty_arg', False)
        self.kwargs['append_arg'] = self.arg
        self.kwargs['request_arg'] = self.request_arg

    def __get_serializers(self, kwargs):
        serializer_class = kwargs.pop('serializer_class', self.view.serializer_class)

        if 'serializer_class_one' in kwargs:
            serializer_class_one = kwargs.pop('serializer_class_one')
        elif hasattr(self.view, 'serializer_class_one'):
            serializer_class_one = self.view.serializer_class_one
        else:
            # This option is deprecated because all viewsets return `serializer_class_one`
            serializer_class_one = serializer_class  # nocv

        return (serializer_class, serializer_class_one)

    def _get_subs_from_view(self) -> _t.Iterable:
        # pylint: disable=protected-access
        extra_acts = map(lambda x: x[0], getmembers(self.view, _is_extra_action))
        filter_subs = self.filter_subs
        return filter(lambda name: name not in filter_subs, extra_acts)

    def get_subs(self) -> _t.Iterable:
        subs = self._get_subs_from_view()

        if self.allowed_subs is None:
            return []
        elif self.allowed_subs:
            return set(self.allowed_subs).intersection(subs)

        return subs

    @property
    def serializer(self):
        return self.serializers[0]

    @property
    def serializer_one(self):
        return self.serializers[-1]

    def get_view(self, name: _t.Text, **options):
        # pylint: disable=redefined-outer-name,too-many-statements
        mixin_class: _t.Type[NestedViewMixin] = NestedViewMixin
        if hasattr(self.view, 'create'):
            if self.kwargs.get('allow_append', False):
                mixin_class = NestedWithAppendMixin
            else:
                mixin_class = NestedWithoutAppendMixin

        tp = name.split('_')[-1]
        if tp == 'detail':
            detail = True
        elif tp == 'list':
            detail = False
        else:
            detail = getattr(self.view, options['nested_sub']).detail

        manager_name = self.kwargs.get('manager_name', self.name)
        view_class = utils.get_if_lazy(self.view)

        class NestedView(mixin_class, view_class):  # type: ignore
            __slots__ = ('nested_detail',)  # pylint: disable=class-variable-slots-conflict
            __doc__ = self.view.__doc__
            format_kwarg = None
            queryset_filters = self.queryset_filters

        NestedView.__name__ = self.view.__name__  # type: ignore
        NestedView.nested_detail = detail  # type: ignore

        def nested_view_function_wrapper(view_obj, request, *args, **kwargs):
            kwargs.update(options)
            view_obj.nested_parent_object = view_obj.get_object()
            nested_append_arg = view_obj.nested_append_arg
            nested_request_arg = view_obj.nested_arg
            nested_parent_object = view_obj.nested_parent_object

            if nested_append_arg:
                nested_id = getattr(nested_parent_object, nested_append_arg, None)
            else:
                nested_id = None

            if callable(manager_name):
                nested_manager = manager_name(nested_parent_object)
            else:
                if hasattr(nested_parent_object, manager_name):
                    nested_manager = getattr(nested_parent_object, manager_name)
                else:
                    view_manager_function_name = f'get_manager_{manager_name}'
                    nested_manager_func = getattr(view_obj, view_manager_function_name)
                    nested_manager = nested_manager_func(nested_parent_object)

            NestedView.__name__ = self.view.__name__
            NestedView.master_view = view_obj
            NestedView.lookup_field = nested_append_arg
            NestedView.lookup_url_kwarg = nested_request_arg
            NestedView.nested_detail = detail
            NestedView.nested_allow_append = view_obj.nested_allow_append
            NestedView.nested_append_arg = nested_append_arg
            NestedView.nested_request_arg = nested_request_arg
            NestedView.nested_parent_object = nested_parent_object
            NestedView.nested_id = nested_id
            NestedView.nested_manager = nested_manager

            getattr(view_obj, 'nested_allow_check', lambda *args, **kwargs: None)()
            return nested_view_function(view_obj, NestedView, request, *args, **kwargs)

        nested_view_function_wrapper.__name__ = name
        nested_view_function_wrapper.__doc__ = self.view.__doc__
        nested_view_function_wrapper._nested_view = self.view  # type: ignore
        nested_view_function_wrapper._nested_wrapped_view = NestedView  # type: ignore
        return name, nested_view_function_wrapper

    def get_view_type(self, type_name: _t.Text, **options):
        return self.get_view(f'{self.name}_{type_name}', **options)

    def get_list_view(self, **options):
        return self.get_view_type('list', **options)

    def get_detail_view(self, **options):
        return self.get_view_type('detail', **options)

    def get_sub_view(self, sub: _t.Text, **options):
        return self.get_view_type(sub, nested_sub=sub, **options)

    def get_decorator(self, detail=False, **options) -> _t.Callable:
        args = [self.name]
        args += [self.arg] if detail else []
        args += self.args
        kwargs = dict(self.kwargs)
        kwargs['methods'] = self.methods
        kwargs['serializer_class'] = self.serializer_one if detail else self.serializer
        kwargs['filterset_class'] = getattr(self.view, 'filterset_class', []) or getattr(self.view, 'filter_class', [])
        kwargs.update(options)
        return nested_action(*args, **kwargs)

    def _filter_methods(self, methods, detail=False):
        allowed_methods = set(self.view.get_view_methods(detail))
        return allowed_methods.intersection(methods)

    def decorated(self, detail) -> _t.Tuple[_t.Text, _t.Type[_t.Union[NestedViewMixin, base.GenericViewSet]]]:
        name, view = self.get_detail_view() if detail else self.get_list_view()
        kwargs = {'detail': detail}
        kwargs['url_name'] = f'{self.name}-{"detail" if detail else "list"}'
        if not detail:
            kwargs['suffix'] = 'List'
        if self.methods:
            kwargs['methods'] = self._filter_methods(self.methods, detail=detail)
        else:
            kwargs['methods'] = self.view.get_view_methods(detail)  # type: ignore
        view_class = self.get_decorator(**kwargs)(view)
        view_class._nested_view = self.view
        view_class._nested_name = self.name
        view_class._nested_subname = self.name
        view_class._nested_wrapped_view = view._nested_wrapped_view
        return name, view_class

    def decorated_list(self):
        return self.decorated(detail=False)

    def decorated_detail(self):
        return self.decorated(detail=True)

    def _get_decorated_sub(self, sub):
        name, subaction_view = self.get_sub_view(sub)
        sub_view = getattr(self.view, sub)
        sub_path = sub_view.url_path
        decorator = self.get_decorator(
            detail=sub_view.detail,
            sub_opts={'sub_path': sub_path},
            methods=sub_view.mapping or self.methods,
            serializer_class=sub_view.kwargs.get('serializer_class', self.serializer),
            url_name=f'{self.name}-{sub_view.url_name}',
            _nested_args=getattr(sub_view, '_nested_args', OrderedDict())
        )
        view = decorator(subaction_view)
        existing_swagger_auto_schema = getattr(view, '_swagger_auto_schema', {})
        view._swagger_auto_schema = getattr(
            sub_view, '_swagger_auto_schema', existing_swagger_auto_schema
        )
        view._nested_view = getattr(sub_view, '_nested_view', self.view)
        view._nested_name = sub
        view._nested_subname = getattr(sub_view, '_nested_subname', self.name)
        view._nested_wrapped_view = getattr(sub_view, '_nested_wrapped_view', None)
        return name, view

    def generate_decorated_subs(self):
        for sub in self._subs:
            yield self._get_decorated_sub(sub)

    def setup(self, view_class: _t.Type[base.GenericViewSet]) -> _t.Type[base.GenericViewSet]:
        if self.arg:
            setattr(view_class, *self.decorated_detail())
            view_class._nested_args = getattr(view_class, '_nested_args', OrderedDict())
            view_class._nested_args[self.name] = self.request_arg
        if self._subs:
            for sub_action_name, sub_action_view in self.generate_decorated_subs():
                setattr(view_class, sub_action_name, sub_action_view)
        setattr(view_class, *self.decorated_list())
        return view_class


def extend_viewset_attribute(name: _t.Text, override: bool = False, data: _t.Optional[_t.Any] = None) -> _t.Callable:
    def wrapper(view_class: _t.Type[base.GenericViewSet]):
        if not override:
            attr_data = tuple(list(getattr(view_class, name, ())) + list(data or ()))
        else:
            attr_data = data  # type: ignore
        setattr(view_class, name, attr_data)
        return view_class

    return wrapper


def extend_filterbackends(backends: _t.Iterable, override: bool = False) -> _t.Callable:
    return extend_viewset_attribute('filter_backends', override, backends)


def cache_method_result(func: _t.Callable) -> _t.Callable:
    """Decorator that caches return value of method based on args and kwargs,
    cache value stored in the object instance"""
    name: _t.Text = f'__cache_{func.__name__}'

    def wrapper(self, *args, **kwargs):
        result = getattr(self, name, None)

        if result is None or result[0] != args or result[1] != kwargs:
            result = (args, kwargs, func(self, *args, **kwargs))
            setattr(self, name, result)

        return result[2]

    return wrapper



import orjson
from django.core.files.uploadedfile import SimpleUploadedFile
from django.core.files.images import ImageFile
from django.db.models import TextField, ForeignKey, FileField, ImageField, Field
from django.db.models.fields.files import FileDescriptor, FieldFile
from django.db.models.query_utils import DeferredAttribute
from vstutils.utils import raise_context_decorator_with_default

"""
These model fields used in :class:`vstutils.api.serializers.VstSerializer`
to form model_field_class: serializer_field_class mapping
"""


class MultipleFieldFile(FieldFile):
    """
    Subclasses :class:`django.db.models.fields.files.FieldFile`. Provides :meth:`MultipleFieldFile.save`
    and :meth:`MultipleFieldFile.delete` to manipulate the underlying file, as well as update the
    associated model instance.
    """
    def __init__(self, instance, field, name):
        super(MultipleFieldFile, self).__init__(instance, field, name)
        self._old_name = self.name

    def _set_attr_value(self):
        """
        Set new value of file to object attr.
        """
        setattr(
            self.instance,
            self.field.attname,
            [
                self.name if file.name == self._old_name else file
                for file in getattr(self.instance, self.field.attname)
            ]
        )

    def _clear_attr_value(self):
        """
        Pop None values from file list.
        """
        setattr(
            self.instance,
            self.field.attname,
            [
                file
                for file in getattr(self.instance, self.field.attname) if file.name != self.name
            ]
        )

    def save(self, name, content, save=True):
        """
        Save changes in file to storage and to object attr.
        """
        self._old_name = self.name
        name = self.field.generate_filename(self.instance, name)
        self.name = self.storage.save(name, content, max_length=self.field.max_length)
        self._set_attr_value()
        self._committed = True

        if save:
            self.instance.save()

    def delete(self, save=True):
        """
        Delete file from storage and from object attr.
        """
        if not self:
            return  # nocv

        if hasattr(self, '_file'):
            self.close()
            del self.file

        self.storage.delete(self.name)

        self._clear_attr_value()
        self._committed = False

        if save:
            self.instance.save()


class MultipleFileDescriptor(FileDescriptor):
    """
    Subclasses :class:`django.db.models.fields.files.FileDescriptor` to handle list of files.
    Return a list of :class:`MultipleFieldFile` when accessed so you can write code like:

    .. sourcecode:: python

        from myapp.models import MyModel
        instance = MyModel.objects.get(pk=1)
        instance.files[0].size
    """
    def get_file(self, file, instance):
        """
        Always return valid attr_class object.For details on logic see
        :meth:`django.db.models.fields.files.FileDescriptor.__get__`.
        """
        if isinstance(file, str) or file is None:
            attr = self.field.attr_class(instance, self.field, file)
            file = attr

        elif isinstance(file, SimpleUploadedFile):
            file_copy = self.field.attr_class(instance, self.field, file.name)
            file_copy.file = file
            file_copy._committed = False  # pylint: disable=W0212 protected-access
            file = file_copy

        elif isinstance(file, MultipleFieldFile) and instance != file.instance:  # nocv
            file.instance = instance
            file.field = self.field
            file.storage = self.field.storage

        return file

    def __get__(self, instance, cls=None):
        """
        Return list of MultipleFieldFile at all times.
        """
        if instance is None:
            return self  # nocv

        instance.__dict__[self.field.attname] = [
            self.get_file(file, instance)
            for file in DeferredAttribute.__get__(self, instance, cls) or []
        ]

        return instance.__dict__[self.field.attname]


class MultipleFileMixin:
    """
    Mixin suited to use with :class:`django.db.models.fields.files.FieldFile` to transform it to
    a Field with list of files.
    """
    def __init__(self, **kwargs):
        kwargs['max_length'] = None
        super().__init__(**kwargs)

    def pre_save(self, model_instance, add):
        """
        Call .save() method on every file in list
        """
        files = getattr(model_instance, self.attname)
        for file in files:
            if file and not file._committed:  # pylint: disable=W0212 protected-access
                file.save(file.name, file.file, save=False)
        return files

    def get_prep_value(self, value):
        """
        Prepare value for database insertion
        """
        value = Field.get_prep_value(self, value)
        if value is None:
            return value

        return orjson.dumps(list(map(str, value))).decode('utf-8')

    @raise_context_decorator_with_default(default=[])
    def from_db_value(self, value, expression, connection):
        """
        Transform db value to an internal value
        """
        if value:
            return orjson.loads(value)
        return value  # nocv

    def get_internal_type(self):
        return "TextField"


class MultipleFileField(MultipleFileMixin, FileField):
    """
    Subclasses :class:`django.db.models.fields.files.FileField`.
    Field for storing a list of Storage-kept files. All args passed to FileField.
    """
    attr_class = MultipleFieldFile
    descriptor_class = MultipleFileDescriptor
    description = "List of Files"


class MultipleImageFieldFile(ImageFile, MultipleFieldFile):
    """
    Subclasses :class:`MultipleFieldFile` and :class:`ImageFile mixin`,
    handles deleting _dimensions_cache when file is deleted.
    """
    def delete(self, save=True):
        if hasattr(self, '_dimensions_cache'):
            del self._dimensions_cache
        super().delete(save)


class MultipleImageField(MultipleFileMixin, ImageField):
    """
    Field for storing a list of storage-kept images. All args are passed to
    :class:`django.db.models.fields.files.ImageField`, except height_field and width_field,
    they are not currently implemented.
    """
    attr_class = MultipleImageFieldFile
    descriptor_class = MultipleFileDescriptor
    description = "List of Images"

    def update_dimension_fields(self, instance, force=False, *args, **kwargs):
        pass


class NamedBinaryFileInJSONField(TextField):
    """
    Extends :class:`django.db.models.TextField`. Use this field in :class:`vstutils.models.BModel` to get
    `vstutils.api.NamedBinaryFileInJSONField` in serializer.
    """


class NamedBinaryImageInJSONField(NamedBinaryFileInJSONField):
    """
    Extends :class:`django.db.models.TextField`. Use this field in :class:`vstutils.models.BModel` to get
    `vstutils.api.NamedBinaryImageInJSONField` in serializer.
    """


class MultipleNamedBinaryFileInJSONField(TextField):
    """
    Extends :class:`django.db.models.TextField`. Use this field in :class:`vstutils.models.BModel` to get
    `vstutils.api.MultipleNamedBinaryFileInJSONField` in serializer.
    """


class MultipleNamedBinaryImageInJSONField(MultipleNamedBinaryFileInJSONField):
    """
    Extends :class:`django.db.models.TextField`. Use this field in :class:`vstutils.models.BModel` to get
    `vstutils.api.MultipleNamedBinaryImageInJSONField in serializer`.
    """


class FkModelField(ForeignKey):
    """
    Extends :class:`django.db.models.ForeignKey`. Use this field in :class:`vstutils.models.BModel` to get
    `vstutils.api.FkModelField in serializer`. To set Foreign Key relation set `to` argument to string path to model
    or to Model Class as in :class:`django.db.models.ForeignKey`
    """



# flake8: noqa: E501
TRANSLATION = {
    'new': 'Mới',
    'Create': 'Tạo',
    'add': 'thêm',
    'Edit': 'Chỉnh sửa',
    'edit': 'Chỉnh sửa',
    'Save': 'lưu',
    'Remove': 'Xóa',
    'Reload': 'Làm mới',
    'send': 'gửi',
    'select': 'chọn',
    'filters': 'bộ lọc',
    'apply': 'áp dụng',
    'cancel': 'Hủy',
    'field': 'tường',
    'actions': 'thao tác',
    'sublinks': 'đường dẫn',
    'name': 'tên',
    'type': 'loại',
    'owner': 'chủ sở hữu',
    'notes': 'ghi chú',
    'copy': 'sao chép',
    'error': 'lỗi',
    'home': 'trang chủ',
    'login': 'đăng nhập',
    'logout': 'đăng xuất',
    'Profile': 'Hồ sơ',
    'settings': 'thiết lập',
    'Settings': 'Thiết lập',
    'version': 'phiên bản',
    'cache': 'bộ nhớ đệm',
    'documentation': 'tài liệu',
    'request': 'yêu cầu',
    'repository': 'kho thông tin',
    'app info': 'thông tin',
    'status': 'trạng thái',
    'List is empty': 'Danh sách trống',
    'child instances': 'thêm các mục lồng',
    'enter value': 'nhập giá trị',
    'more info': 'thêm thông tin',
    'search by': 'tìm kiếm theo',
    'Search': 'Tìm kiếm ',
    'Selected': 'Đã chọn',
    'Nothing found!': 'Không tìm thấy gì!',
    'Main settings': 'cài đặt chính',
    'Custom settings': 'Cài đặt tùy chỉnh',
    'Remove from list': 'Xóa từ danh sách',
    'Purge': 'Xóa',
    'Select an action for this instances': 'Chọn một hành động cho trường hợp này',
    'Username not': 'Tên người dùng không phải',

    # fields
    'Id': 'định danh',
    'ID': 'định danh',

    # x-menu generic
    'system': 'Hệ thống',
    'users': 'Người dùng',
    'Is active': 'Hoạt động',
    'Is staff': 'Người điều hành',

    # multiactions button
    'actions on': 'thao tác trên',
    'on item': 'thành phần',

    'key': 'Chìa khóa',
    'value': 'Giá trị',
    'new owner': 'Chủ sở hữu mới',
    'detail': 'Chi tiết',
    'description': 'Miêu tả',
    'username': 'Tên người dùng',
    'is active': 'Hoạt động',
    'user': 'Người dùng',
    'yes': 'Có',
    'no': 'Không',
    'is staff': 'Admin',
    'first name': 'Tên',
    'last name': 'Họ',
    'password': 'Mật khẩu',
    'Password': 'Mật khẩu',
    'Change password': 'Đổi mật khẩu',
    'Change_password': 'Đổi mật khẩu',
    'generate password': 'Viết mật khẩu',
    'repeat password': 'Lập lại mật khẩu',
    'Repeat password': 'Lập lại mật khẩu',
    'Old password': 'Mật khẩu cũ',
    'New password': 'Mật khẩu mới',
    'Confirm new password': 'Xác nhận mật khẩu mới',
    'kind': 'Loại',
    'mode': 'Chế độ',
    'options': 'Tùy chọn',
    'information': 'Thông tin',
    'date': 'Ngày',

    'file n selected': 'File chưa được chọn',
    'image n selected': 'Hình ảnh chưa được chọn',

    # filters
    'Ordering': 'Thứ tự phân loại',
    'is_active': 'Hoạt động',
    'Username': 'Tên đăng nhập',
    'First name': 'Tên',
    'Last name': 'Họ',

    'Page matching current url was not found': 'Không tìm thấy url hiện tại phù hợp với trang',
    'Unknown error': 'Lỗi không thể nhận biết',
    'Error': 'Lỗi',

    'We\'ve emailed you instructions for setting your password, if an account exists with the email you entered. You should receive them shortly.': 'Hướng dẫn hủy mật khẩu đã được gửi vào e-mail của bạn.',
    'If you don\'t receive an email, please make sure you\'ve entered the address you registered with, and check your spam folder.': '"Nếu bạn chưa nhận được thư, hãy kiểm tra lại địa chỉ e-mail và mục Spam."',
    'The password reset link was invalid, possibly because it has already been used. Please request a new Password reset.': '"Đường dẫn không đúng, có thể đã được sử dụng. Xin hãy thử lại lần nữa."',
    'Please enter your new password twice so we can verify you typed it in correctly.': 'Nhập mật khẩu mới',
    'change my password': 'Thay đổi mật khẩu',
    'Your password has been set.  You may go ahead and log in now.': 'Thay đổi mật khẩu thành công. Hãy đăng nhập.',
    'Please enter a correct %(username)s and password. Note that both fields may be case-sensitive.': 'Vui lòng nhập chính xác% (tên người dùng) s và mật khẩu',
    'or': 'hoặc',
    'two-factor authentication code': 'hai bước xác thực (2FA)',
    'verify code': 'xác nhận mã',
    'Enter the code from the two-factor app on your mobile device. If you\'ve lost your device, you may enter one of your recovery codes.': '"Nhập mã từ ứng dụng sử dụng hai bước xác thực trên thiết bị di động của bạn.Nếu bạn mất thiết bị của mình, bạn có thể nhập một trong các mã khôi phục."',
    'Scan the image with the two-factor authentication app on your phone. If you can’t use a barcode, enter this text code instead.': '"Quét hình ảnh bằng ứng dụng xác thực hai bước trên điện thoại của bạn. Nếu bạn không thể quét hình ảnh, hãy nhập mã văn bản."',
    'Enter the six-digit code from the application': 'Điền mã xác thực từ ứng dụng',
    'code': 'Mã',
    'recovery codes': 'Mã khôi phục',
    'Recovery codes are used to access your account in the event you cannot receive two-factor authentication codes.': 'Mã khôi phục được sử dụng để truy cập vào tài khoản của bạn nếu bạn không thể nhận được mã xác thực hai bước.',
    'Disabling Two-Factor Authentication (2FA) will decrease your account\'s security': 'mã xác thực hai bước.',
    'Enable': 'Bật',
    'Disable': 'Tắt',
    'Two factor authentication': 'Xác thực hai bước (2FA)',
    'No {0} matches the given query.': 'Không tìm thấy',
    'Password is not correct.': 'Sai mật khẩu',
    'New passwords values are not equal.': 'Mật khẩu mới không chính xác',
    'This password is entirely numeric.': 'Mật khẩu này chỉ có chữ số.',
    'Passwords do not match.': 'Mật khẩu không phù hợp.',
    "The two password fields didn't match.": 'Hai trường mật khẩu không khớp với nhau.',
    'Dark theme': 'Màn hình tối',
    'Language': 'Ngôn ngữ',
    'height': 'chiều cao',
    'width': 'chiều rộng',
    'More info': 'thêm thông tin',
    'Dark mode': 'Chủ đề tối',
    'Execute': 'Hành hình',
    'Domain should be setuped or username should be with @domain.name': "Tên miền phải được thiết lập hoặc tên người dùng phải bằng @ domain.name",
    'Invalid name': 'Tên không hợp lệ',
    'Should be [full.domain.name]': 'Phải là [full.domain.name].',
    'Invalid registration email send.': 'Email đăng ký không hợp lệ gửi.',
    'To continue, need to accept the terms agreement.': 'Để tiếp tục, cần chấp nhận thỏa thuận điều khoản.',

    # Проверить
    'Changes in settings are successfully saved. Please refresh the page.': 'Các thay đổi trong cài đặt đã được lưu thành công. Vui lòng làm mới trang.',
    'now': 'hiện nay',
    'later': 'một lát sau',
    'I accept the ': 'Tôi chấp nhận ',
    'Zoom': 'Thu phóng',
    'Scale': 'Tỉ lệ',
    'This file format is not supported': 'Định dạng tệp này không được hỗ trợ',
    'Oops! Something went wrong.': 'Một cái gì đó đã đi sai.',
    '[Object not found]': '[Không tìm thấy đối tượng]',
    'Execute action on {0}': 'Thực hiện hành động trên {0}',
    '{n} instance': '{n} trường hợp',
    'Open image': 'Mở hình ảnh',
    'Confirm action': 'Xác nhận hành động',
    'Confirm': 'Xác nhận',
    'Cancel': 'Hủy bỏ',

    # field's validation
    # empty
    'Field "<b>{0}</b>" is empty.': 'Trường "<b> {0} </b>" trống không',
    # required
    'Field "<b>{0}</b>" is required.': 'Trường "<b> {0} </b>"  bắt buộc.',
    # minLength
    'Field "<b>{0}</b>" is too short.<br> Field length should not be shorter, then {1}.': 'Độ dài của trường "<b> {0} </b>" quá ngắn. <br> Độ dài của trường không được nhỏ hơn {1}.',
    # maxLength
    'Field "<b>{0}</b>" is too long. <br> Field length should not be longer, then {1}.': 'Độ dài của trường "<b> {0} </b>" quá dài. <br> Trường không được dài hơn {1}.',
    # min
    'Field "<b>{0}</b>" is too small.<br> Field should not be smaller, then {1}.': 'Giá trị của trường "<b> {0} </b>" quá nhỏ. <br> Giá trị của trường không được nhỏ hơn {1}.',
    # max
    'Field "<b>{0}</b>" is too big.<br> Field should not be bigger, then {1}.': 'Giá trị của trường "<b> {0} </b>" quá lớn. <br> Giá trị của trường không được lớn hơn {1}.',
    # invalid
    '<b>{0} </b> value is not valid for <b>{1}</b> field.': 'Giá trị <b> {0} </b> không hợp lệ cho trường <b> {1} </b>.',
    # email
    '<b>"{0}"</b> field should be written in <b>"example@mail.com"</b> format.': 'Trường <b> "{0}" </b> phải làm như <b> "example@mail.com" </b>.',

    # validators
    'imageValidationResolutionError': 'Phải có ít nhất {min} và không nhiều hơn {max} pixel.',

    # instance operation success
    # add
    'Child "<b>{0}</b>" instance was successfully added to parent list.': 'Đối tượng con "<b> {0} </b>" đã được thêm thành công vào danh sách mẹ.',
    # create
    'New "<b>{0}</b>" instance was successfully created.': 'Đối tượng mới "<b> {0} </b>" đã được tạo thành công.',
    # remove
    '"<b>{0}</b>" {1} was successfully removed.': '"<b>{0}</b>" {1} đã được xóa thành công."',
    # save
    'Changes in "<b>{0}</b>" {1} were successfully saved.': 'Thay đổi đối tượng {1} "<b> {0} </ b>" đã được lưu thành công.',
    # execute
    'Action "<b>{0}</b>" was successfully executed on "<b>{1}</b>" instance.': 'Thao tác "<b> {0} </b>" đã được thực hiện thành công trên đối tượng "<b> {1} </b>".',
    # instance operation error
    # add:
    'An error occurred during adding of child "<b>{0}</b>" instance to parent list.<br> Error details: {1}': 'Khi thêm đối tượng con "<b> {0} </b>" vào danh sách mẹ đã xảy ra lỗi <br> thêm chi tiết: {1}',
    # create:
    'An error occurred during creation.<br>Error details:<br>{0}': 'Đã xảy ra một số lỗi khi tạo phiên bản mới. <br> Chi tiết lỗi: <br> {0}',
    # remove:
    'An error occurred during removal process of "<b>{0}</b>" {1}.<br> Error details: {2}': 'Khi xóa {1} "<b> {0} </b>". đã xảy ra lỗi <br> thêm chi tiết: {2}',
    # save:
    'An error occurred during process.<br>Error details:<br>{0}': 'Đã xảy ra một số lỗi trong quá trình lưu. <br> Chi tiết lỗi: <br> {0}',
    # execute:
    'An error occurred during <b>{0}</b>. Error details:<br>{1}': 'Đã xảy ra một số lỗi trong <b> {0} </b>. Chi tiết lỗi: <br> {1}',
    'Link': 'Liên kết',

    # csv
    'Actions': 'Hành động',
    'Index': 'Mục lục',
    'Download with headers': 'Tải xuống với tiêu đề',
}

SERVER_TRANSLATION = {
    # field's validation
    "value is not a valid list": "giá trị không phải là một danh sách hợp lệ",
    "value is not a valid JSON": "giá trị không phải là JSON hợp lệ",
    "key {missing_key} is missing": "khóa {missing_key} bị thiếu",
    "invalid key {invalid_key}": "khóa không hợp lệ {invalid_key}",

    # field help_text
    'A confirmation will be sent to your e-mail': 'Một xác nhận sẽ được gửi đến e-mail của bạn',
    'Required. Inform a valid email address': 'Yêu cầu. Thông báo một địa chỉ email hợp lệ',

    # validators
    'unsupported image file format, expected ({0}), got {1}': 'Định dạng hình ảnh không được hỗ trợ, Được hỗ trợ ({0}), đã nhận {1}',
    'For some reason, this image file cannot be opened': 'Không thể mở file hình ảnh',
    'Invalid image {0}. Expected from {1} to {2}, got {3}': 'Hình ảnh {0} không chính xác. Dự kiến ​​từ {1} đến {2}, đã nhận {3}',
    'This field must contain only digits but input: ': 'Trường này chỉ được chứa các chữ số nhưng nhập: ',
    'Unsupported image file format': 'Định dạng tệp hình ảnh không được hỗ trợ',
    'is not in listed supported types': 'không có trong các loại được hỗ trợ được liệt kê',
    'Invalid image size orientations': 'Kích thước hình ảnh không đúng',
    'Current image size': 'Kích thước hình ảnh hiện tại',

    # agreements
    'personal data processing policy': 'Chính sách xử lý dữ liệu cá nhân',
    'I agree with ': 'Tôi đồng ý với ',
    'the personal data processing policy': 'chính sách xử lý dữ liệu cá nhân',
    'To continue, need to agree to the personal data processing policy.': 'Để tiếp tục, cần đồng ý với chính '
                                                                          'sách xử lý dữ liệu cá nhân.',
    'terms of agreement': 'các điều khoản của thỏa thuận',

    # auth/reg
    'sign in to start your session': 'đăng nhập để bắt đầu',
    'confirm your email before logging in': 'xác nhận email của bạn trước khi đăng nhập',
    'Confirmation link is invalid or expired': 'Liên kết xác nhận không hợp lệ hoặc hết hạn',
    'sign in': 'đăng nhập',
    'sign up': 'đăng ký',
    'forgot password': 'quên mật khẩu',
    'confirm password': 'xác nhận mật khẩu',
    'register': 'đăng ký',
    'fields with * is required': 'các mục có dấu * là bắt buộc',
    'Forgot your password? Enter your email address below, and an email with instructions for setting a new one will be sent.': 'Quên mật khẩu? Nhập E-mail để nhận hướng dẫn khôi phục mật khẩu',
    'reset my Password': 'đặt lại mật khẩu của tôi',
    'Invalid authentication code': 'Mã xác thực không hợp lệ',
    'Secret string must be provided': 'Chuỗi bí mật phải được cung cấp',

    # filters help text
    'A unique integer value (or comma separated list) identifying this instance.': 'Giá trị số duy nhất',
    'A name string value (or comma separated list) of instance.': 'Tên đối tượng',
    'Which field to use when ordering the results.': 'Trường để sắp xếp kết quả',
    "Search by {}'s primary key or {}": 'Tìm kiếm theo khóa chính của {} hoặc {}',
    '_name': 'tên',
    'Boolean value meaning status of user.': 'Giá trị boolean cho biết trạng thái của người dùng',
    'Users first name.': 'Tên người dùng',
    'Users last name.': 'Họ người dùng',
    'Users e-mail value.': 'E-mail người dùng',
}



import json

from django import template, forms
from django.core.exceptions import NON_FIELD_ERRORS

from ..utils import translate as _

register = template.Library()


@register.inclusion_tag('vst_inclusion_tags/b64_img_from_json_string.html')
def img_from_json(json_string, **kwargs):
    media_type = kwargs.get('media_type', 'image/png')
    tag_classes = kwargs.get('tag_classes', '')
    content_attribute = kwargs.get('content_attribute', 'content')

    try:
        content = json.loads(json_string)[content_attribute]
    except Exception:
        content = ''
    return {
        'media_type': media_type,
        'tag_classes': tag_classes,
        'content': content
    }


def _handle_form_fields(form):
    for name, field in form.fields.items():
        field = form[name]
        errors = field.errors
        label = field.label
        classes = []
        label_classes = []
        label_first = True
        hidden = False
        if hasattr(field.field, 'get_bootstrap_label'):
            label = field.field.get_bootstrap_label(field)
        else:
            label = _(label)
        if isinstance(field.field, forms.BooleanField):
            classes.append('form-check')
            label_classes.append('form-check-label')
            label_first = False
        if isinstance(field.field.widget, forms.widgets.HiddenInput):
            hidden = True
        field.field.widget.attrs['class'] = 'form-control' + (field.field.widget.attrs.get('class', '') or '')
        yield {
            'name': name,
            'label': label,
            'bound': field,
            'errors': errors,
            'classes': ' '.join(classes),
            'label_classes': ' '.join(label_classes),
            'label_first': label_first,
            'label_last': not label_first,
            'hidden': hidden,
        }


@register.inclusion_tag('vst_inclusion_tags/bootstrap_form.html')
def bootstrap_form(form: forms.Form, **kwargs):
    return {
        'fields': _handle_form_fields(form),
        'form': form,
        'form_errors': form.errors.get(NON_FIELD_ERRORS, '')
    }



from django.contrib.auth.models import AbstractUser
from django.conf import settings
from rest_framework import permissions

from ..utils import raise_context


class IsAuthenticatedOpenApiRequest(permissions.IsAuthenticated):
    __slots__ = ()

    def is_openapi(self, request):
        return (
            request.path.startswith(f'/{settings.API_URL}/openapi/') or
            request.path.startswith(f'/{settings.API_URL}/endpoint/') or
            request.path == f'/{settings.API_URL}/{request.version}/_openapi/'
        )

    def has_permission(self, request, view):
        return self.is_openapi(request) or super().has_permission(request, view)


class SuperUserPermission(IsAuthenticatedOpenApiRequest):
    __slots__ = ()

    def has_permission(self, request, view):
        if request.user.is_staff or request.method in permissions.SAFE_METHODS:
            # pylint: disable=bad-super-call
            return super(IsAuthenticatedOpenApiRequest, self).has_permission(request, view)
        with raise_context():
            return (
                issubclass(view.get_queryset().model, AbstractUser) and
                str(view.kwargs['pk']) in (str(request.user.pk), 'profile')
            )
        return self.is_openapi(request)

    def has_object_permission(self, request, view, obj):
        if request.user.is_superuser:
            return True
        elif isinstance(obj, AbstractUser) and obj == request.user:
            return True
        return False


class StaffPermission(permissions.IsAdminUser):
    __slots__ = ()



from collections import OrderedDict

from rest_framework.pagination import LimitOffsetPagination as DRFLimitOffsetPagination

from .responses import BaseResponseClass


class LimitOffsetPagination(DRFLimitOffsetPagination):
    def paginate_queryset(self, queryset, request, view=None):
        self.identifier = request.headers.get('Identifiers-List-Name')
        return super().paginate_queryset(queryset, request, view)

    def get_paginated_response(self, data):
        response = BaseResponseClass(OrderedDict([
            ('count', self.count),
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),
            ('results', data)
        ]))
        if self.identifier is not None:
            response.headers['Pagination-Identifiers'] = ','.join(map(str, (d[self.identifier] for d in data)))
        return response



from typing import Text
from collections import OrderedDict as odict
import traceback
import logging
import json

import ldap
from django.conf import settings

from .utils import translate


def json_default(obj):  # nocv
    error_obj = TypeError(f"{type(obj)} is not JSON serializable")
    try:
        if isinstance(obj, bytes):
            return obj.decode("utf-8")
        raise error_obj
    except Exception as err:
        raise error_obj from err


class LDAP:
    # pylint: disable=no-member
    __slots__ = (
        'settings',
        'logger',
        'connection_string',
        'username',
        'password',
        'domain',
        '__conn',
        'user_format',
    )
    fields = ['cn', 'sAMAccountName', 'accountExpires', 'name', 'memberOf']
    LdapError = ldap.LDAPError

    class NotAuth(ldap.INVALID_CREDENTIALS):
        pass

    class InvalidDomainName(ldap.INVALID_CREDENTIALS):
        pass

    def __init__(self, connection_string: Text, username: Text = '', password: Text = None, domain: Text = None):
        """
        LDAP constructor

        :param connection_string: LDAP connection string ('ldap://server')
        :param username: username with domain ('user@domain.name')
                         or without but domain arg should be set.
        :param password: auth password
        :param domain: domain for easy use users
        """
        self.settings = settings
        self.user_format = settings.LDAP_FORMAT.replace('<', "{").replace('>', '}')
        self.logger = logging.getLogger(settings.VST_PROJECT_LIB)
        self.connection_string = connection_string
        self.username = username
        self.password = password
        if domain:
            self.domain = domain
        else:
            self.domain = username.split('@')[-1]
        self.domain = self.domain.strip()
        if not self.domain:
            if r'@' not in username:
                raise self.InvalidDomainName(
                    translate(
                        "Domain should be setuped or username should be with @domain.name"
                    )
                )
        if len(self.domain.split('.')) <= 1:
            raise self.InvalidDomainName(
                translate("Invalid name") +
                f" {domain}. " +
                translate("Should be [full.domain.name].")
            )
        self.auth(self.username, self.password)

    def auth(self, username: Text = None, password: Text = None) -> None:
        self.__conn = self.__authenticate(
            self.connection_string,
            str(username or self.username),
            str(password or self.password)
        )

    def __prepare_user_with_domain(self, username: Text) -> Text:
        user = str(username).split('@', maxsplit=1)[0]
        domain = str(self.domain)
        if len(str(username).split('@')) > 1:
            domain = str(username).split('@', maxsplit=1)[-1]
        domain = domain.lower()
        if domain != user:
            domain = ','.join([f'dc={d}' for d in domain.split('.') if d])
        user = self.user_format.format(username=user, domain=domain)
        self.logger.debug(f'Trying auth in ldap with user "{user}"')
        return user

    def __authenticate(self, ad: Text, username: Text, password: Text) -> ldap.functions.LDAPObject:
        """
        Active Directory auth function

        :param ad: LDAP connection string ('ldap://server')
        :param username: username with domain ('user@domain.name')
        :param password: auth password
        :return: ldap connection or None if error
        """
        result = None
        conn = ldap.initialize(ad)
        conn.protocol_version = 3
        conn.set_option(ldap.OPT_REFERRALS, 0)
        user = self.__prepare_user_with_domain(username)
        self.logger.debug(f"Trying to auth with user '{user}' to {ad}")
        try:
            conn.simple_bind_s(user, password)
            result = conn
            self.username, self.password = username, password
            self.logger.debug(f"Successfull login as {username}")
        except ldap.INVALID_CREDENTIALS:
            result = False
            self.logger.debug(traceback.format_exc())
            self.logger.debug("Invalid ldap-creds.")
        except Exception as ex:  # nocv
            self.logger.debug(traceback.format_exc())
            self.logger.debug(f"Unknown error: {str(ex)}")

        return result

    def __get_user_data(self):
        data_list = self.username.split("@")
        if len(data_list) < 2:
            return self.username, self.domain
        return data_list

    @property
    def domain_user(self) -> Text:
        return self.__get_user_data()[0].split('\\')[-1]

    @property
    def domain_name(self) -> Text:
        return self.__get_user_data()[1]

    def isAuth(self) -> bool:
        """
        Indicates that object auth worked
        :return: True or False
        """
        if isinstance(self.__conn, ldap.ldapobject.LDAPObject) or self.__conn:
            return True
        return False

    def __ldap_filter(self, *filters):
        dc_list = [f"dc={i}" for i in self.domain_name.split('.') if i]
        additinal_filter = "".join([f"({i})" for i in filters if i])
        s_filter = f'(&(objectCategory=user){additinal_filter})'
        base_dn = f"{','.join(dc_list)}"
        self.logger.debug(
            f'Search in LDAP: {json.dumps(odict(BASE_DN=base_dn, FILTER=s_filter, FIELDS=self.fields))}'
        )
        return base_dn, ldap.SCOPE_SUBTREE, s_filter, self.fields

    def group_list(self, *args) -> Text:
        if not self.isAuth():
            raise self.NotAuth("Invalid auth.")
        try:
            data = {
                k: v for k, v in self.__conn.search_s(*self.__ldap_filter(*args)) if k
            }
            return json.dumps(data, indent=4, ensure_ascii=False, default=json_default)
        except Exception:  # nocv
            self.logger.debug(traceback.format_exc())
            raise

    def __repr__(self):  # nocv
        return str(self)

    def __unicode__(self):  # nocv
        return str(self)

    def __str__(self):  # nocv
        msg = 'authorized' if self.isAuth() else "unauthorized"
        return f'[ {msg} {self.connection_string} -> {self.username} ]'

    def __del__(self):
        if isinstance(getattr(self, '__conn', None), ldap.ldapobject.LDAPObject):
            self.__conn.unbind_s()  # nocv



from configparserc import tools


def multikeysort(items, columns, reverse=False):
    if not isinstance(items, list):
        items = list(items)  # nocv
    if not isinstance(columns, list):
        columns = list(columns)  # nocv
    columns.reverse()

    for column in columns:
        # pylint: disable=cell-var-from-loop
        is_reverse = column.startswith('-')
        if is_reverse:
            column = column[1:]
        items.sort(key=lambda row: row[column], reverse=is_reverse)

    if reverse:
        items.reverse()

    return items


get_file_value = tools.get_file_value
File = tools.File



# pylint: disable=import-error
from collections import OrderedDict

import django
from django.conf import settings
from django.urls import reverse_lazy
from rest_framework import __version__ as drf_version  # type: ignore
from drf_yasg import openapi, __version__ as drf_yasg_version  # type: ignore

from ... import __version__ as vstutils_version


x_versions = OrderedDict()  # type: OrderedDict
x_versions['application'] = settings.PROJECT_VERSION
x_versions['library'] = settings.PROJECT_LIB_VERSION
x_versions['vstutils'] = vstutils_version
x_versions['django'] = django.__version__
x_versions['djangorestframework'] = drf_version
x_versions['drf_yasg'] = drf_yasg_version

api_info_dict = OrderedDict(
    title=settings.PROJECT_GUI_NAME,
    default_version=settings.VST_API_VERSION,
    description=settings.SWAGGER_API_DESCRIPTION,
    terms_of_service=settings.TERMS_URL,
    contact=openapi.Contact(**settings.CONTACT),
    **{
        'x-versions': x_versions,
        'x-links': settings.OPENAPI_EXTRA_LINKS,
        'x-menu': settings.PROJECT_GUI_MENU,
        'x-docs': OrderedDict(
            has_docs='docs' in settings.INSTALLED_APPS,
            docs_url=getattr(settings, 'DOC_URL', '/docs/'),
        ),
        'x-settings': OrderedDict(
            static_path=settings.STATIC_URL,
            enable_gravatar=settings.ENABLE_GRAVATAR,
            time_zone=settings.TIME_ZONE,
            logout_url=reverse_lazy('logout'),
            login_url=reverse_lazy('login'),
        ),
        'x-page-limit': settings.PAGE_LIMIT,
    }
)

api_info = openapi.Info(**api_info_dict)



import os
import time
import copy

import jwt
from rest_framework import request as drf_request
from django.conf import settings
from drf_yasg import generators
from drf_yasg.inspectors import field as field_insp
from vstutils.utils import import_class, raise_context_decorator_with_default


def get_centrifugo_public_address(request: drf_request.Request):
    address = settings.CENTRIFUGO_PUBLIC_HOST
    if address.startswith('/'):
        address = request.build_absolute_uri(address)
    elif 'api' in address.rsplit('/', 2):
        address = address[:address.rfind('api')]
    return os.path.join(address.replace('http', 'ws', 1), 'connection/websocket')


class EndpointEnumerator(generators.EndpointEnumerator):
    api_url_prifix = f'^{settings.API_URL}/'

    def get_api_endpoints(self, *args, **kwargs):
        if 'prefix' in kwargs and (not kwargs['prefix'].startswith(self.api_url_prifix)):
            return []
        return super().get_api_endpoints(*args, **kwargs)


class VSTSchemaGenerator(generators.OpenAPISchemaGenerator):
    endpoint_enumerator_class = EndpointEnumerator

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.version:
            self.version = settings.VST_API_VERSION

    def _get_hooks(self):
        return map(
            raise_context_decorator_with_default(),
            filter(
                bool,
                map(
                    raise_context_decorator_with_default()(import_class),
                    getattr(settings, 'OPENAPI_HOOKS', ())
                )
            )
        )

    def _get_subname(self, name):
        names = name.split('_')
        return ['_'.join(names[0:-1]), names[-1]]

    def _get_model_type_info(self, name, model, model_field=None, query_name=None):
        if model_field:
            return None
        _query_name, field_name = self._get_subname(name)
        query_name = query_name or _query_name
        related_model = field_insp.get_related_model(model, query_name)
        related_field = field_insp.get_model_field(related_model, field_name)
        if related_field is None:
            return None
        type_info = field_insp.get_basic_type_info(related_field)
        if type_info is None:
            return None  # nocv
        type_info.update({
            'description':
                ('A unique {} value identifying '
                 'instance of this {} sublist.').format(type_info['type'], query_name),
        })
        return type_info

    def _get_manager_name(self, param, view_cls):
        name, _ = self._get_subname(param['name'])
        if not hasattr(view_cls, f'{name}_detail'):
            return None
        sub_view = getattr(view_cls, f'{name}_detail', None)
        return getattr(sub_view, '_nested_manager', None)

    def _update_param_model(self, param, model, model_field=None, **kw):
        type_info = self._get_model_type_info(param['name'], model, model_field, **kw)
        if type_info is None:
            return type_info
        param.update(type_info)
        return param

    def _update_param_view(self, param, model, view_cls):
        manager_name = self._get_manager_name(param, view_cls)
        if manager_name is None:
            return None
        return self._update_param_model(param, model, query_name=manager_name)

    def get_path_parameters(self, path, view_cls):
        parameters = super().get_path_parameters(path, view_cls)
        queryset = getattr(view_cls, 'queryset', None)
        for param in parameters:
            model, model_field = field_insp.get_queryset_field(queryset, param['name'])
            if self._update_param_model(param, model, model_field):
                continue
            elif self._update_param_view(param, model, view_cls):
                continue
        return parameters

    def get_operation_keys(self, subpath, method, view):
        keys = super().get_operation_keys(subpath, method, view)
        subpath_keys = list(filter(bool, subpath.split('/')))
        r_type, gist = keys[-1], keys[-2]
        if method.upper() == 'GET' and '_detail' in r_type:
            keys = keys[:-1] + ['_'.join(r_type.split('_')[:-1])] + ['get']
        if r_type == 'get' and subpath_keys[-1] == gist:
            if any(f for f in dir(view) if f.endswith('_'.join([gist, 'list']))):
                keys[-1] = 'list'
        return keys

    def get_schema(self, request: drf_request.Request = None, *args, **kwargs):  # type: ignore
        # pylint: disable=signature-differs
        if not getattr(request, 'version', ''):
            request.version = self.version  # type: ignore
        result = super().get_schema(request, *args, **kwargs)
        if request and getattr(request, 'accepted_media_type', None) == 'application/openapi+json':
            result['info']['x-user-id'] = request.user.pk
            if settings.CENTRIFUGO_CLIENT_KWARGS and request.user.is_authenticated:
                secret = settings.CENTRIFUGO_CLIENT_KWARGS.get('token_hmac_secret_key', '')
                if secret and request.user.pk:
                    result['info']['x-centrifugo-token'] = jwt.encode(
                        {
                            "sub": request.session.session_key,
                            "exp": int(time.time() + request.session.get_expiry_age()),
                            "info": {
                                'user_id': request.user.pk
                            }
                        },
                        secret,
                        algorithm="HS256"
                    )
                    result['info']['x-centrifugo-address'] = get_centrifugo_public_address(request)

        for hook in self._get_hooks():
            result = copy.deepcopy(result)
            hook(request=request, schema=result)

        return result



from ..api import decorators as api_decorators, serializers


class register_view_decorator:  # pylint: disable=invalid-name
    __slots__ = ('method_type', 'args', 'kwargs', 'inherit')

    def __init__(self, method_type, *args, **kwargs):
        self.method_type = f'type_{method_type}'
        assert hasattr(self, self.method_type), f'Invalid register type {method_type}.'
        self.args = args
        self.kwargs = kwargs
        self.inherit = kwargs.pop('inherit', False)

    def type_action(self, func):
        if func.__doc__ and 'description' not in self.kwargs:
            self.kwargs['description'] = func.__doc__
        if 'response_serializer' not in self.kwargs and 'serializer_class' not in self.kwargs:
            self.kwargs['serializer_class'] = serializers.EmptySerializer
        return api_decorators.subaction(*self.args, **self.kwargs)(func)

    def type_override_method(self, func):
        return func

    def __call__(self, func):
        result = getattr(self, self.method_type)(func)
        result._append_to_view = True
        result._inherit = self.inherit
        return result


class register_view_action(register_view_decorator):  # pylint: disable=invalid-name
    """
    Decorator for turning model methods to generated view
    `actions <https://www.django-rest-framework.org/api-guide/viewsets/#marking-extra-actions-for-routing>`_.
    The decorated method becomes a method of generated view and `self` is an view object.
    See supported args in :func:`vstutils.api.decorators.subaction`.

    .. note::
        Sometimes you may need to use proxy models with a common set of actions.
        To receive the action by the proxy model, pass the named argument ``inherit`` with ``True`` value.

    .. note::
        Often, an action does not transfer any parameters and requires only sending an empty query.
        To speed up development, we set the default serializer to :class:`vstutils.api.serializers.EmptySerializer`.

    """
    __slots__ = ()  # type: ignore

    def __init__(self, *args, **kwargs):
        super().__init__('action', *args, **kwargs)


class register_view_method(register_view_decorator):  # pylint: disable=invalid-name
    __slots__ = ()  # type: ignore

    def __init__(self, *args, **kwargs):
        super().__init__('override_method', *args, **kwargs)



import base64
import typing as _t
import re
import warnings
from io import BytesIO
from mimetypes import guess_type

from django.conf import settings
from rest_framework import serializers
from vstutils.utils import raise_context, translate

try:
    from PIL import Image, UnidentifiedImageError

    has_pillow = True
except ImportError:  # nocv
    has_pillow = False


def get_orientations_with_comma(*args):
    for arg in '.,;'.join(args).split(';'):
        yield from arg.split('.')


def apply_ratio(apply_to, ratio=1, rounded=False):
    return {
        side: rounded and round(side_size * ratio) or side_size * ratio
        for side, side_size in apply_to.items()
    }


def find_broken_orientation(size, limits) -> _t.Tuple:
    return tuple(
        str(orient)
        for orient, orient_limits in limits.items()
        if not (orient_limits['min'] <= size[orient] <= orient_limits['max'])
    )


def resize_image_from_to(img, limits):
    """
    Utility function to resize image proportional to values between min and max values for each side.
    Can create white margins if it's needed to satisfy restrictions

    :param img: Pillow Image object
    :type img: PIL.Image
    :param limits: Dict with min/max side restrictions like:
                   ``{'width': {'min': 300, 'max: 600'}, 'height':  {'min': 400, 'max: 800'}}``
    :type limits: dict
    :return: Pillow Image object
    :rtype: PIL.Image
    """
    size = {'width': img.size[0], 'height': img.size[1]}
    limits_list = [{'orient': k, **v} for k, v in limits.items()]

    best_size = apply_ratio(
        size,
        ratio=limits_list[0]['max'] / size[limits_list[0]['orient']]
    )
    # decrease next side if needed
    if len(limits_list) == 2:
        ratio = limits_list[1]['max'] / best_size[limits_list[1]['orient']]
        if ratio <= 1:
            best_size = apply_ratio(best_size, ratio=ratio)

    best_size = apply_ratio(best_size, rounded=True)

    resized_image = img.resize((best_size['width'], best_size['height']), Image.NEAREST)
    broken_orientation = next(iter(find_broken_orientation(best_size, limits)), None)
    if not broken_orientation:
        return resized_image
    # add margin to image
    summary_margin_size = limits[broken_orientation]['min'] - best_size[broken_orientation]
    margin_parts = [int(summary_margin_size / 2), summary_margin_size - int(summary_margin_size / 2)]
    margin_sides = broken_orientation == "width" and ['left', 'right'] or ['top', 'bottom']
    margin = {margin_sides[i]: margin_parts[i] for i in range(2)}
    margined_image = Image.new(
        resized_image.mode,
        (resized_image.size[0] + margin.get('left', 0) + margin.get('right', 0),
         resized_image.size[1] + margin.get('top', 0) + margin.get('bottom', 0)),
        (255, 255, 255)
    )
    margined_image.paste(resized_image, (margin.get('left', 0), margin.get('top', 0)))
    return margined_image


def resize_image(img, width, height):
    """
    Utility function to resize image proportional to specific values.
    Can create white margins if it's needed to satisfy required size

    :param img: Pillow Image object
    :type img: PIL.Image
    :param width: Required width
    :type width: int
    :param height: Required height
    :type height: int
    :return: Pillow Image object
    :rtype: PIL.Image
    """
    return resize_image_from_to(img, {
        'width': {'min': width, 'max': width},
        'height': {'min': height, 'max': height},
    })


class RegularExpressionValidator:
    """
    Class for regular expression based validation

    :raises rest_framework.exceptions.ValidationError: in case value does not match regular expression
    """
    regexp: _t.Optional[_t.Pattern[_t.Text]] = None

    def __init__(self, regexp: _t.Optional[_t.Pattern] = None):
        """
        :param regexp: Compiled regular expression
        """
        self._regexp = regexp if regexp else self.regexp

    def __call__(self, value):
        if value and not self._regexp.match(value):
            raise serializers.ValidationError(self._error_msg())

    def _error_msg(self):
        return translate('This field must match pattern ') + f'{self._regexp.pattern}'


class UrlQueryStringValidator(RegularExpressionValidator):
    """Class for validation url query string, for example a=&b=1"""

    regexp = re.compile(r'^[^&?=].+=.*$')


class ImageValidator:
    """
    Base Image Validation class
    Validates image format
    Won't work if Pillow isn't installed
    :param extensions: Tuple or List of file extensions, that should pass the validation
    Raises rest_framework.exceptions.ValidationError: in case file extension are not in the list
    """
    warning_msg = 'Warning! Pillow is not installed, validation has not been done'
    default_extensions: _t.Union[_t.Tuple, _t.List] = [
        'bmp',
        'jpeg',
        'jpg',
        'png',
    ]

    def __init__(self, extensions: _t.Optional[_t.Union[_t.Tuple, _t.List]] = None, **kwargs):
        if extensions is not None:
            assert isinstance(extensions, (tuple, list)), "extensions must be list or tuple"
        else:
            extensions = self.default_extensions
        self.extensions = tuple(sorted(  # type: ignore
            filter(
                bool,
                {
                    guess_type(f'name.{e}')[0]
                    for e in extensions
                }
            )
        ))
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __call__(self, value):
        if not self.has_pillow:
            warnings.warn(self.warning_msg, ImportWarning)
            return
        file_media_type = guess_type(value['name'])[0]
        if value and self.extensions and file_media_type not in self.extensions:
            raise serializers.ValidationError(
                translate('Unsupported image file format') +
                f' "{value["name"]}" ({file_media_type or ""}) ' +
                translate('is not in listed supported types') +
                f' ({",".join(self.extensions)}).'
            )
        self.media_type = file_media_type.split(sep='/')[1].upper()

    @property
    def has_pillow(self):
        """
        Check if Pillow is installed
        """
        return has_pillow


class ImageOpenValidator(ImageValidator):
    """
    Image validator that checks if image can be unpacked from b64 to PIL Image obj

    Raises rest_framework.exceptions.ValidationError if PIL throws error when trying to open image
    """
    error_msg = 'For some reason, this image file cannot be opened'

    def __call__(self, value):
        if not self.has_pillow:
            warnings.warn(self.warning_msg, ImportWarning)
            return
        super().__call__(value)
        try:
            self.img = Image.open(BytesIO(base64.b64decode(value['content'])))
        except UnidentifiedImageError as err:
            raise serializers.ValidationError(self.error_msg) from err


class ImageBaseSizeValidator(ImageOpenValidator):
    """
    Validates image size
    To use this class for validating image width/height, rewrite
    self.orientation to ('height',) or ('width',) or ('height', 'width')

    Raises rest_framework.exceptions.ValidationError: if not(min <= (height or width) <= max)
    """
    orientation: _t.Union[_t.Union[_t.Tuple[str], _t.Tuple[str, str]], _t.Tuple] = ()

    requires_context = True

    def __call__(self, value, serializer_field=None):
        if not self.has_pillow:
            warnings.warn(self.warning_msg, ImportWarning)
            return
        super().__call__(value)
        should_resize_image = False
        with raise_context():
            should_resize_image = serializer_field.context['request'].headers.get('Auto-Resize-Image', None) == 'true'
        self.validate(value, should_resize_image)

    def validate(self, img_data, should_resize_image):
        limits = {}
        size = {}
        for orientation in self.orientation:
            min_value = getattr(self, f'min_{orientation}', 1)
            max_value = getattr(self, f'max_{orientation}', float('inf'))
            value = getattr(self.img, orientation)
            limits[orientation] = {'min': min_value, 'max': max_value}
            size[orientation] = value

        broken_orientations = find_broken_orientation(size, limits)
        if broken_orientations:
            if should_resize_image and settings.ALLOW_AUTO_IMAGE_RESIZE:
                buffered = BytesIO()
                resize_image_from_to(self.img, limits).save(buffered, format=self.media_type)
                img_data['content'] = base64.b64encode(buffered.getvalue()).decode('utf-8')
            else:
                raise serializers.ValidationError([
                    'Invalid image size orientations', ': ',
                    *tuple(get_orientations_with_comma(*broken_orientations)),
                    '. ',
                    'Current image size', ': ',
                    'x'.join([str(size[x]) for x in broken_orientations]),
                ])


class ImageHeightValidator(ImageBaseSizeValidator):
    """
    Wrapper for ImageBaseSizeValidator that validates only height

    :param min_height: minimal height of an image being validated
    :param max_height: maximal height of an image being validated
    """
    orientation = ('height',)


class ImageWidthValidator(ImageBaseSizeValidator):
    """
    Wrapper for ImageBaseSizeValidator that validates only width

    :param min_width: minimal width of an image being validated
    :param max_width: maximal width of an image being validated
    """
    orientation = ('width',)


class ImageResolutionValidator(ImageBaseSizeValidator):
    """
    Wrapper for ImageBaseSizeValidator that validates both height and width

    :param min_height: minimal height of an image being validated
    :param max_height: maximal height of an image being validated
    :param min_width: minimal width of an image being validated
    :param max_width: maximal width of an image being validated
    """
    orientation = ('height', 'width')



# pylint: disable=no-member,no-classmethod-decorator,protected-access
import operator
import uuid
from functools import lru_cache, partial, reduce
from itertools import chain
from copy import deepcopy

from django_filters import rest_framework as filters, filterset
from django.core.cache import caches as django_caches
from django.db.models.base import ModelBase
from django.db.models import fields as django_model_fields
from django.db.models.fields.related import ManyToManyField, OneToOneField, ForeignKey
from django.utils.functional import SimpleLazyObject, lazy
from django.db.models.signals import post_save, post_delete
from django.dispatch import receiver
from rest_framework.fields import ModelField, JSONField, CharField as drfCharField, ChoiceField
from rest_framework.mixins import CreateModelMixin, UpdateModelMixin, DestroyModelMixin
from vstutils.utils import translate as _

from ..api.fields import (
    NamedBinaryFileInJsonField,
    NamedBinaryImageInJsonField,
    MultipleNamedBinaryFileInJsonField,
    MultipleNamedBinaryImageInJsonField,
    FkField,
    PasswordField,
    RelatedListField,
)
from ..utils import (
    import_class,
    apply_decorators,
    classproperty,
    get_if_lazy,
    raise_context_decorator_with_default,
)
from ..api import (
    base as api_base,
    filters as api_filters,
    serializers as api_serializers,
    decorators as api_decorators,
)
from ..api.filter_backends import DeepViewFilterBackend

# Constants
LAZY_MODEL = object()
DEFAULT_VIEW_FIELD_NAMES = (
    'name',
    'title',
    'username',
    'email',
    'key',
)
EXCLUDED_FIELDS = (
    NamedBinaryFileInJsonField,
    NamedBinaryImageInJsonField,
    MultipleNamedBinaryFileInJsonField,
    MultipleNamedBinaryImageInJsonField,
    ModelField,
    JSONField,
    PasswordField,
    RelatedListField,
)
CHANGE_MIXINS = (
    CreateModelMixin,
    UpdateModelMixin,
    DestroyModelMixin
)

view_settings = (
    'filter_backends',
    'permission_classes',
    'authentication_classes',
    'throttle_classes',
    'renderer_classes',
    'parser_classes',
)

default_extra_metadata: dict = {
    # list or class which is base for view
    "view_class": None,
    # base class for serializers
    "serializer_class": None,
    # name of openapi model
    "serializer_class_name": None,
    # tuple or list of fields in list view
    "list_fields": None,
    # dict which override fields types of list view serializer
    "override_list_fields": None,
    # tuple or list of fields in detail view
    "detail_fields": None,
    # dict which override fields types of detail view serializer
    "override_detail_fields": None,
    # name of default view field
    "view_field_name": None,
    # list or tuple of non-bulk methods from gui
    "non_bulk_methods": None,
    # dict which indicates about properties groups
    "properties_groups": None,
    # key-value of actions serializers (key - action, value - serializer class)
    "extra_serializer_classes": None,
    # tuple or list of filters on list
    "filterset_fields": 'serializer',
    # tuple or list of fields using for search requests
    "search_fields": None,
    # additional attrs which means that this view allowed to copy elements
    "copy_attrs": None,
    # key-value mapping with nested views (key - nested name, kwargs for nested decorator)
    "nested": None,
    # additional view's settings
    **dict(reduce(operator.add, [((f'pre_{i}', None), (i, None), (f'override_{i}', False)) for i in view_settings]))
}


# Handlers
def _translate_search_help_text(field_name, related_name):
    return _("Search by {}'s primary key or {}").format(_(field_name), _(related_name))


def is_append_to_view(item):
    return bool(getattr(item[1], '_append_to_view', False))


def is_inherit_append_to_view(item):
    return bool(getattr(item[1], '_inherit', False)) and is_append_to_view(item)


def get_append_to_view(items, handler=is_append_to_view):
    return dict(filter(handler, items))


def get_parents_append_to_view(mro_items):
    result = {}
    for items in reversed(list(mro_items)):
        result.update(get_append_to_view(items, is_inherit_append_to_view))
    return result


def update_cache_for_model(instance, **kwargs):
    kwargs.get('cached_model_class', instance.__class__).set_etag_value()


def get_first_match_name(field_names, default=None):
    return next(
        (i for i in field_names if i in DEFAULT_VIEW_FIELD_NAMES),
        next(iter(field_names[1:]), default)
    )


def _get_unicode(obj):
    return obj.__unicode__()


def _ensure_pk_in_fields(model_class, fields):
    if fields is None or fields == '__all__':
        return
    fields = list(fields)
    primary_key_name = model_class._meta.pk.attname
    if primary_key_name not in fields:
        fields.insert(0, primary_key_name)
    return fields


def _import_class_if_string(value):
    if isinstance(value, str):
        return SimpleLazyObject(lambda: import_class(value))
    return value


def _load_settings_for_view(metadataobject):
    return list(map(_import_class_if_string, metadataobject))


def _get_setting_for_view(metatype, metadata, views):
    override = metadata[f'override_{metatype}']
    metadataobject = metadata[metatype] or []
    if metadataobject:
        metadataobject = _load_settings_for_view(metadataobject)
    pre_metadataobject = metadata[f'pre_{metatype}'] or []
    if override:
        return pre_metadataobject + metadataobject  # nocv
    if metadataobject or pre_metadataobject:
        for view in views:
            if hasattr(view, metatype):  # pragma: no branch
                return list(chain(pre_metadataobject, getattr(view, metatype), metadataobject))
        return pre_metadataobject + metadataobject  # nocv


def _result_with_arg_decorator(func):
    def wrapper(*args, **kwargs):
        return (func(*args, **kwargs), *args)

    return wrapper


def _bool_first(item):
    return item and item[0]


def _get_decorator(data):
    path, deco_kwargs = data

    assert not ('model' in deco_kwargs and 'view' in deco_kwargs), (
        "Invalid model configuration: "
        f"Unable to set 'model' and 'view' at the same time for path [{path}]."
    )
    if 'model' in deco_kwargs:
        model = _import_class_if_string(deco_kwargs.pop('model'))
        assert isinstance(model, ModelBaseClass), (
            f"Invalid model type {type(model)} for path [{path}]."
        )
        deco_kwargs['view'] = model.lazy_generated_view
        if 'arg' not in deco_kwargs:
            deco_kwargs['arg'] = model._meta.pk.name
    else:
        deco_kwargs['view'] = _import_class_if_string(deco_kwargs.pop('view'))
    return api_decorators.nested_view(path, **deco_kwargs)


def get_proxy_labels(model):
    labels = []
    if model._meta.proxy:
        labels.append(model._meta.proxy_for_model._meta.label)
        labels += get_proxy_labels(model._meta.proxy_for_model)
    return labels


# Classes
class ApplyNestedDecorators(apply_decorators):
    def __init__(self, nested: dict):
        super().__init__(
            *map(_get_decorator, nested.items())
        )


class ModelBaseClass(ModelBase, metaclass=classproperty.meta):
    """
    Metaclass for all models. Read more in :class:`vstutils.models.BModel`
    """

    def __new__(mcs, name, bases, attrs, **kwargs):
        if "__slots__" not in attrs:
            attrs['__slots__'] = ()
        if "__unicode__" in attrs and '__str__' not in attrs:
            attrs['__str__'] = _get_unicode
        extra_metadata: dict = {**default_extra_metadata}
        meta = None
        if "Meta" in attrs:
            meta = attrs['Meta']
            extra_metadata['proxy'] = getattr(meta, 'proxy', False)
            if not getattr(meta, 'abstract', False):
                for extra_name in filter(lambda y: hasattr(meta, y), map(lambda x: f'_{x}', extra_metadata.keys())):
                    extra_metadata[extra_name[1:]] = getattr(meta, extra_name)
        attrs['__extra_metadata__'] = deepcopy(extra_metadata)
        model_class = super(ModelBaseClass, mcs).__new__(mcs, name, bases, attrs, **kwargs)
        model_class.OriginalMeta = meta if meta is not None else model_class.Meta
        if hasattr(model_class, '__prepare_model__'):
            model_class.__prepare_model__()
        if getattr(model_class, '_cache_responses', False):
            receiver(post_save, sender=model_class)(update_cache_for_model)
            receiver(post_delete, sender=model_class)(update_cache_for_model)
            cache_related_labels = getattr(model_class, '_cache_related_labels', ())
            update_cache_for_model_related = partial(update_cache_for_model, cached_model_class=model_class)
            for label in cache_related_labels:
                receiver(post_save, sender=label)(update_cache_for_model_related)
                receiver(post_delete, sender=label)(update_cache_for_model_related)
        return model_class

    def get_proxy_labels(cls):
        return get_proxy_labels(cls)  # nocv

    def get_api_cache_name(cls):
        return f'api_caching_table_{cls._meta.db_table}'

    def get_etag_value(cls):
        # pylint: disable=no-value-for-parameter
        return str(django_caches['etag'].get_or_set(cls.get_api_cache_name(), str(uuid.uuid4())))

    def set_etag_value(cls):
        # pylint: disable=no-value-for-parameter
        new_value = str(uuid.uuid4())
        django_caches['etag'].set(cls.get_api_cache_name(), new_value)
        return new_value

    @classproperty
    @lru_cache()
    def generated_view(cls):
        # pylint: disable=no-value-for-parameter
        return cls.get_view_class()

    @classproperty
    @lru_cache()
    def lazy_generated_view(cls, **extra_metadata):
        # pylint: disable=unnecessary-lambda,no-value-for-parameter
        return SimpleLazyObject(lambda: cls.get_view_class(**extra_metadata))

    def get_model_fields_mapping(cls, filter_handler=lambda f: True, fields_getter=None):
        fields_getter = fields_getter if fields_getter is not None else lambda: cls._meta.fields
        return {
            f.name: f
            for f in fields_getter()
            if filter_handler(f)
        }

    def get_all_view_fields_lazy(cls):
        return SimpleLazyObject(
            lambda: tuple(cls.get_model_fields_mapping(
                lambda f: not isinstance(f, (ManyToManyField, OneToOneField)) and not getattr(f, '_hide', False)
            ).keys())
        )

    def get_serializer_class(  # noqa: CFQ002
            cls,
            serializer_class,
            serializer_class_name=None,
            fields=None,
            field_overrides=None,
            view_field_name=None,
            non_bulk_methods=None,
            metadata=None,
    ):
        # pylint: disable=no-value-for-parameter,too-many-arguments
        attributes = {}

        if view_field_name:
            attributes['_view_field_name'] = view_field_name

        if serializer_class is None:
            serializer_class = api_serializers.VSTSerializer

        serializer_class = _import_class_if_string(serializer_class)

        serializer_class_name = cls.__name__ + 'Serializer' if serializer_class_name is None else serializer_class_name

        if fields and fields != '__all__':
            fields = list(fields)
        else:
            fields = cls.get_all_view_fields_lazy()

        meta = type('Meta', (), {
            'model': cls,
            'ref_name': serializer_class_name.replace('Serializer', ''),
            'fields': fields
        })

        serializer_class = get_if_lazy(serializer_class)
        properties_groups = (metadata if isinstance(metadata, dict) else cls.get_extra_metadata())['properties_groups']
        translate_model = getattr(cls, '_translate_model', None)

        if properties_groups:
            attributes['_schema_properties_groups'] = dict(**properties_groups)

        if non_bulk_methods:
            attributes['_non_bulk_methods'] = non_bulk_methods

        if translate_model:
            attributes['_translate_model'] = translate_model

        return type(serializer_class)(
            serializer_class_name,
            (serializer_class,),
            {
                "Meta": meta,
                **attributes,
                **(field_overrides or {})
            }
        )

    def _update_serializers(cls, metadata: dict, serializers: dict):
        """
        Setup extra serializers.
        """
        for serializer_name, extra_serializer_class in (metadata['extra_serializer_classes'] or {}).items():
            if issubclass(extra_serializer_class, api_serializers.VSTSerializer) and \
                    getattr(extra_serializer_class.Meta, 'model', None) is None:
                inject_from = getattr(extra_serializer_class.Meta, '__inject_from__', None)
                field_overrides = {
                    n: f
                    for n, f in (metadata.get(f'override_{inject_from}_fields', {}) or {}).items()
                    if extra_serializer_class._declared_fields.get(n, None) is None
                } or None
                extra_serializer_class = cls.get_serializer_class(  # pylint: disable=no-value-for-parameter
                    serializer_class=extra_serializer_class,
                    serializer_class_name=extra_serializer_class.__name__,
                    fields=_ensure_pk_in_fields(
                        cls,
                        getattr(extra_serializer_class.Meta, 'fields', None) or
                        metadata.get(f'{inject_from}_fields', None)
                    ),
                    field_overrides=field_overrides,
                    view_field_name=(
                        metadata['view_field_name']
                        if inject_from is not None
                        else getattr(extra_serializer_class, '_view_field_name', None)
                    ),
                    non_bulk_methods=(
                        getattr(extra_serializer_class, '_non_bulk_methods', None) or metadata['non_bulk_methods']
                        if inject_from is not None and metadata['non_bulk_methods']
                        else getattr(extra_serializer_class, '_non_bulk_methods', None)
                    ),
                    metadata=metadata,
                )
            elif issubclass(extra_serializer_class, api_serializers.VSTSerializer) and \
                    getattr(extra_serializer_class.Meta, 'model', None) is LAZY_MODEL:
                meta_class = type(extra_serializer_class.Meta)(
                    'Meta',
                    (extra_serializer_class.Meta,),
                    {'model': cls}
                )
                extra_serializer_class = type(extra_serializer_class)(
                    extra_serializer_class.__name__,
                    (extra_serializer_class,),
                    {'Meta': meta_class}
                )
            serializers[serializer_name] = extra_serializer_class

    def get_extra_metadata(cls):
        return cls.__extra_metadata__.copy()

    def get_list_serializer_name(cls, metadata=None):
        # pylint: disable=no-value-for-parameter
        serializer_class_name = (metadata or cls.get_extra_metadata())['serializer_class_name']
        if serializer_class_name is None:
            serializer_class_name = cls.__name__ + 'Serializer'
        return serializer_class_name

    @raise_context_decorator_with_default(default=None)
    def _get_filterset_class(cls, filterset_fields, serializers):
        """
        Logic with generating filterset class. Returns `None` if class is not generated by any reason.
        """
        if filterset_fields == 'serializer':
            serializer = serializers['serializer_class']
            filterset_fields = serializer.Meta.fields

            if not isinstance(filterset_fields, str):  # pragma: no branch
                model_fields = tuple(cls.get_model_fields_mapping().keys())  # pylint: disable=no-value-for-parameter
                # fields are added to tuple if field in model_fields and not in EXCLUDED_FIELDS
                filterset_fields = tuple(
                    f for f in filterset_fields if f in model_fields and
                    not isinstance(serializer().fields.get(f), EXCLUDED_FIELDS)
                )

        if filterset_fields:

            if isinstance(filterset_fields, dict):
                filterset_fields_list = tuple(filterset_fields.keys())
                filterset_fields_types = {k: v for k, v in filterset_fields.items() if v is not None}
            else:
                filterset_fields_list, filterset_fields_types = filterset_fields, {}

            class Meta:
                model = cls if not cls._meta.proxy else cls._meta.proxy_for_model
                fields = filterset_fields_list

            filter_base_classes = []
            if 'id' in filterset_fields_list:
                filter_base_classes.append(api_filters.DefaultIDFilter)
            if 'name' in filterset_fields_list:
                filter_base_classes.append(api_filters.DefaultNameFilter)
            if not filter_base_classes:
                filter_base_classes.append(filters.FilterSet)

            iteration_filter_handler = (
                lambda f: f.name not in filterset_fields_types and f.name in filterset_fields_list
            )
            serializer = serializers['serializer_class']()
            for field_name, field in cls.get_model_fields_mapping(iteration_filter_handler).items():
                if isinstance(field, ForeignKey):
                    related_name = get_first_match_name([
                        f.name
                        for f in field.related_model._meta.fields
                        if isinstance(f, django_model_fields.CharField)
                    ])
                    filterset_fields_types[field_name] = api_filters.CharFilter(
                        method=api_filters.FkFilterHandler(
                            related_pk=field.target_field.attname,
                            related_name=related_name
                        ),
                        help_text=lazy(_translate_search_help_text)(field_name, related_name)
                    )
                elif isinstance(field, django_model_fields.BooleanField):
                    filterset_fields_types[field_name] = filters.BooleanFilter()
                elif isinstance(serializer.fields.get(field_name), ChoiceField):
                    filterset_fields_types[field_name] = filters.ChoiceFilter(
                        choices=tuple(serializer.fields.get(field_name).choices.items())
                    )

            return filterset.FilterSetMetaclass(
                f'{cls.__name__}FilterSetClass',
                tuple(filter_base_classes),
                {'Meta': Meta, **filterset_fields_types}
            )

    def _get_search_fields(cls, serializer, fields=None):
        if fields is not None:
            yield from fields
        else:
            model_fields = cls.get_model_fields_mapping()  # pylint: disable=no-value-for-parameter
            serializer_fields = serializer().fields
            avail_fields = filter(
                tuple(model_fields.keys()).__contains__,
                tuple(
                    k if v.source is None or v.source == '*' else v.source
                    for k, v in serializer_fields.items()
                    if not isinstance(v, EXCLUDED_FIELDS)
                )
            )
            for field in avail_fields:
                serializer_field = serializer_fields[field]
                if isinstance(serializer_field, FkField):
                    yield f'{field}__{serializer_field.autocomplete_represent}'
                elif isinstance(serializer_field, drfCharField):
                    yield field

    def _get_view_class(cls, view_base_class):
        """
        Get one item of view base class for inheritance.
        """
        if view_base_class is None:
            return api_base.ModelViewSet
        elif view_base_class == 'read_only':
            return api_base.ReadOnlyModelViewSet
        elif view_base_class == 'list_only':
            return api_base.ListOnlyModelViewSet
        elif view_base_class == 'history':
            return api_base.HistoryModelViewSet  # nocv
        elif isinstance(view_base_class, str):
            return import_class(view_base_class)
        return view_base_class

    def get_view_class(cls, **extra_options):  # noqa: CFQ001
        """
        Method which return autogenerated ViewSet based on model's Meta class.
        """
        # pylint: disable=no-value-for-parameter
        metadata = cls.get_extra_metadata()
        metadata.update(extra_options)
        list_fields = _ensure_pk_in_fields(cls, metadata['list_fields'])
        detail_fields = _ensure_pk_in_fields(cls, metadata['detail_fields'] or list_fields)

        view_attributes = {'model': cls, **metadata.get('extra_view_attributes', {})}

        serializer_class = metadata['serializer_class']
        serializers = {
            'serializer_class': cls.get_serializer_class(  # pylint: disable=no-value-for-parameter
                serializer_class=serializer_class,
                serializer_class_name=cls.get_list_serializer_name(metadata),  # pylint: disable=no-value-for-parameter
                fields=list_fields,
                field_overrides=metadata['override_list_fields'] or {},
                view_field_name=metadata['view_field_name'],
                non_bulk_methods=metadata['non_bulk_methods'],
                metadata=metadata,
            )
        }
        detail_fields_override = metadata['override_detail_fields']
        if not detail_fields_override and not metadata['detail_fields']:
            detail_fields_override = metadata['override_list_fields']

        serializers['serializer_class_one'] = cls.get_serializer_class(  # pylint: disable=no-value-for-parameter
            serializer_class=serializer_class,
            serializer_class_name=f'One{serializers["serializer_class"].__name__}',
            fields=detail_fields,
            field_overrides=detail_fields_override or {},
            view_field_name=metadata['view_field_name'],
            non_bulk_methods=metadata['non_bulk_methods'],
            metadata=metadata,
        )
        cls._update_serializers(metadata, serializers)

        view_class_data = metadata['view_class']

        if not isinstance(view_class_data, (tuple, list)):
            view_class_data = (view_class_data,)

        view_class = [cls._get_view_class(v) for v in view_class_data]
        if getattr(cls, '_cache_responses', False):
            view_class.insert(0, api_base.CachableHeadMixin)

        if metadata['copy_attrs']:
            view_attributes.update(map(lambda r: (f'copy_{r[0]}', r[1]), metadata['copy_attrs'].items()))
            view_class.append(api_base.CopyMixin)

        filterset_class = cls._get_filterset_class(metadata['filterset_fields'], serializers)
        if filterset_class is not None:
            view_attributes['filterset_class'] = filterset_class

        if hasattr(cls, 'deep_parent_field'):
            metadata['pre_filter_backends'] = metadata['pre_filter_backends'] or []
            metadata['pre_filter_backends'].insert(0, DeepViewFilterBackend)

        get_setting_for_view = partial(
            _result_with_arg_decorator(_get_setting_for_view),
            metadata=metadata,
            views=view_class
        )
        for value, name in filter(_bool_first, map(get_setting_for_view, view_settings)):
            view_attributes[name] = SimpleLazyObject(lambda obj=value: list(map(get_if_lazy, obj)))

        view_attributes['search_fields'] = SimpleLazyObject(lambda: tuple(cls._get_search_fields(
            serializers['serializer_class_one'],
            metadata['search_fields']
        )))

        generated_view = type(
            f'{cls.__name__}ViewSet',
            tuple(view_class),
            {
                **view_attributes,
                **serializers,
                **get_append_to_view(vars(cls).items()),
                **get_parents_append_to_view([vars(c).items() for c in cls.mro()]),
            }
        )

        generated_view = apply_decorators(
            *map(_import_class_if_string, getattr(cls, 'generated_view_decorators', []))
        )(ApplyNestedDecorators(metadata['nested'] or {})(generated_view))

        if hasattr(cls, 'deep_parent_field') and issubclass(generated_view, CHANGE_MIXINS):
            remote_name: str = cls.get_model_fields_mapping(
                fields_getter=cls._meta.get_fields
            )[cls.deep_parent_field].remote_field.related_name

            parent_view = type(
                f'Parent{cls.__name__}ViewSet',
                (generated_view,),
                {'deep_nested_subview': remote_name}
            )
            return ApplyNestedDecorators({remote_name: {
                'view': generated_view,
                'arg': cls._meta.pk.attname,
                'manager_name': remote_name,
                'allow_append': getattr(cls, 'deep_parent_allow_append', False),
            }})(parent_view)

        return generated_view



# Generated by Django 3.2.3 on 2021-08-13 05:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vstutils_api', '0005_db_translations'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='usersettings',
            options={'default_related_name': '_settings'},
        ),
        migrations.RemoveField(
            model_name='usersettings',
            name='hidden',
        ),
        migrations.AlterField(
            model_name='usersettings',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
    ]



import yaml


Dumper = getattr(yaml, 'CSafeDumper', yaml.SafeDumper)
Loader = getattr(yaml, 'CSafeLoader', yaml.SafeLoader)


class YamlSessionSerializer:
    """
    Simple wrapper around json and yaml to be used in signing.dumps and
    signing.loads.
    """
    __slots__ = ()

    def dumps(self, obj):
        return yaml.dump(obj, Dumper=Dumper).encode('latin-1')

    def loads(self, data):
        return yaml.load(data.decode('latin-1'), Loader=Loader)  # nosec



from functools import wraps, WRAPPER_ASSIGNMENTS
from smtplib import SMTPException

from celery import Celery
from celery.exceptions import Reject
from celery.app.task import BaseTask
from celery.result import AsyncResult
from django.conf import settings
from django.apps import apps

from .utils import import_class, send_template_email_handler, Lock, translate as _

celery_app: Celery = import_class(
    settings.WORKER_OPTIONS['app'].replace(':', '.')  # type: ignore
)


class TaskMeta(type):
    def __new__(mcs, name, bases, attrs, uniq=None):
        task_class = super(TaskMeta, mcs).__new__(mcs, name, bases, attrs)
        task_class.run = mcs.get_notificator_decorator(task_class.run)

        if uniq:
            task_class.run = mcs.get_unique_decorator(task_class.run)
            task_class.apply_async = mcs.get_unique_decorator(task_class.apply_async, is_apply=True)
        elif hasattr(task_class.run, '__uniq_wrapped__') and uniq is not None:
            task_class.run = task_class.run.__uniq_wrapped__
            task_class.apply_async = task_class.apply_async.__uniq_wrapped__

        return task_class

    @staticmethod
    def get_notificator_decorator(func):
        if getattr(func, '__notify_wrapped__', False):
            return func

        @wraps(func, assigned=WRAPPER_ASSIGNMENTS+('__notify_wrapped__',))
        def wrapper(self, *args, **kwargs):
            notifier = (
                getattr(self, '__notifier__', None) or
                apps.get_app_config('vstutils_api').module.notificator_class([])
            )
            with notifier:
                self.__notifier__ = notifier
                result = func(self, *args, **kwargs)
            self.__notifier__ = None
            return result

        wrapper.__notify_wrapped__ = True
        return wrapper

    @staticmethod
    def get_unique_decorator(func, is_apply=False):
        if getattr(func, '__uniq_wrapped__', False):
            return func

        @wraps(func, assigned=WRAPPER_ASSIGNMENTS+('__uniq_wrapped__',))
        def wrapper(self, *args, **kwargs):
            # pylint: disable=protected-access
            if not is_apply and self._get_app().conf.task_always_eager:
                return func(self, *args, **kwargs)
            try:
                with Lock(
                    f'uniq-celery-task-{self.name}',
                    err_msg=_("This task is currently performed by another worker.")
                ):
                    return func(self, *args, **kwargs)
            except Lock.AcquireLockException as err:
                raise Reject(str(err), requeue=False) from err

        wrapper.__uniq_wrapped__ = func
        return wrapper


class TaskClass(BaseTask, metaclass=TaskMeta):
    """
    Wrapper for Celery BaseTask class. Usage is same as Celery standard class, but you can execute task without
    creating instance with :meth:`TaskClass.do` method.

    Example:

        .. sourcecode:: python

            from vstutils.environment import get_celery_app
            from vstutils.tasks import TaskClass

            app = get_celery_app()

            class Foo(TaskClass):
                def run(*args, **kwargs):
                    return 'Foo task has been executed'

            app.register_task(Foo())

    Now you can call your task with various methods:
        - by executing ``Foo.do(*args, **kwargs)``
        - get registered task instance like that - app.tasks['full_path.to.task.class.Foo']

    Also you can make you registered task periodic, by adding it to CELERY_BEAT_SCHEDULE in settings.py:

        .. sourcecode:: python

            CELERY_BEAT_SCHEDULE = {
                'foo-execute-every-month': {
                    'task': 'full_path.to.task.class.Foo',
                    'schedule': crontab(day_of_month=1),
                },
            }

    """

    # pylint: disable=abstract-method

    @property
    def name(self):
        """
        property for proper Celery task execution, needed for :meth:`TaskClass.do` method to work
        """
        return self.__name__

    @property
    def __name__(self):
        return f'{self.__class__.__module__}.{self.__class__.__name__}'

    @classmethod
    def do(cls, *args, **kwargs) -> AsyncResult:
        """
        Method which send signal to celery for start remote task execution.
        All arguments will passed to the task :meth:`TaskClass.run` method.
        """
        return cls().delay(*args, **kwargs)


class SendEmailMessage(TaskClass):
    """
    Task for sending bulk emails, all args and kwargs are passed to :func:`vstutils.utils.send_template_email_handler`.
    Usually you don't need to call this manually, this is called by :func:`vstutils.utils.send_template_email`.
    """
    ignore_result = True

    def run(self, *args, **kwargs):
        try:
            send_template_email_handler(*args, **kwargs)
        except SMTPException as exc:
            raise self.retry(
                exc=exc,
                max_retries=settings.SEND_EMAIL_RETRIES,
                countdown=settings.SEND_MESSAGE_RETRY_DELAY
            )


celery_app.register_task(SendEmailMessage())  # type: ignore



import typing as _t
from pathlib import Path
from django.db.models.query import ModelIterable
from django.db.models.fields import CharField as cf, TextField as tf, IntegerField as intf, BooleanField as bf
from .models.queryset import BQuerySet
from .models.model import BaseModel


class CharField(cf):
    ...


class TextField(tf):
    ...


class IntegerField(intf):
    ...


class BooleanField(bf):
    ...


class CustomQuerySet(BQuerySet):
    custom_iterable_class: _t.ClassVar[_t.Type['CustomModelIterable']]
    custom_query_class: _t.ClassVar[_t.Type['Query']]

    def _filter_or_exclude(self, is_exclude, *args, **kwargs) -> 'CustomQuerySet':
        ...

    def last(self) -> _t.Union[BaseModel, None]:
        ...

    def first(self) -> _t.Union[BaseModel, None]:
        ...

    def setup_custom_queryset_kwargs(self, **kwargs: _t.Any) -> 'CustomQuerySet':
        ...


class CustomModelIterable(ModelIterable):
    def values_handler(self, unit: _t.Dict) -> _t.Dict:
        ...

    def construct_instance(self, data: _t.Dict, model: _t.Type[BaseModel]) -> BaseModel:
        ...

    def __iter__(self) -> _t.Iterator[BaseModel]:
        ...


class ListModel(BaseModel):
    data: _t.ClassVar[_t.List[_t.Dict]]

    @classmethod
    def _get_data(self, chunked_fetch: bool = False, data_source: _t.Iterable = None) -> _t.List[_t.Dict]:
        ...

    def get_pk_value(self) -> _t.Any:
        ...

    def set_pk_value(self, value: _t.Any) -> None:
        ...


class FileModel(ListModel):
    file_path: _t.ClassVar[_t.Union[_t.Text, Path]]

    @classmethod
    def load_file_data(cls) -> _t.Text:
        ...


class Query(_t.Dict):
    distinct_fields: _t.ClassVar[bool]
    queryset: CustomQuerySet
    standard_ordering: _t.Union[bool, property]
    model: _t.Union[ListModel, property]

    def __init__(self, queryset: CustomQuerySet, *args, **kwargs):
        ...

    def chain(self):
        ...

    def clone(self) -> Query:
        ...

    def _check_data(self, check_type: _t.Text, data: _t.Dict):
        ...

    def check_in_query(self, data: _t.Dict):
        ...

    def set_empty(self) -> _t.NoReturn:
        ...

    def set_limits(self, low: int = None, high: int = None):
        ...

    def has_results(self, *args, **kwargs) -> bool:
        ...

    def get_count(self, using) -> int:
        ...

    def can_filter(self) -> bool:
        ...

    def clear_ordering(self, *args, **kwargs) -> None:
        ...

    def add_ordering(self, *ordering) -> None:
        ...



# Generated by Django 3.2 on 2021-06-03 06:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vstutils_api', '0004_user_settings'),
    ]

    operations = [
        migrations.CreateModel(
            name='CustomTranslations',
            fields=[
                ('original', models.CharField(max_length=380, primary_key=True, serialize=False)),
                ('translated', models.CharField(max_length=380)),
                ('code', models.CharField(max_length=5)),
            ],
            options={
                'default_related_name': 'custom_translations',
            },
        ),
        migrations.AddIndex(
            model_name='customtranslations',
            index=models.Index(fields=['original', 'translated', 'code'], name='vstutils_api_translations'),
        ),
    ]



#  pylint: disable=invalid-name
"""
WSGI config for VST Projects project.

It exposes the WSGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/3.2/howto/deployment/wsgi/
"""

import gc
import os
import atexit

from django.core.wsgi import get_wsgi_application

from .environment import prepare_environment

prepare_environment()

# Instagram recommendations.
# gc.disable() doesn't work, because some random 3rd-party library will
# enable it back implicitly.
gc.set_threshold(0)
# Suicide immediately after other atexit functions finishes.
# CPython will do a bunch of cleanups in Py_Finalize which
# will again cause Copy-on-Write, including a final GC
atexit.register(os._exit, 0)  # pylint: disable=protected-access

application = get_wsgi_application()



from typing import Text, Iterable, Callable, Dict, Tuple

from django.db import connections
from rest_framework import status as st

from ..utils import BaseVstObject, import_class


class BaseBackend(BaseVstObject):
    __slots__ = ('__health_methods',)

    def __init__(self):
        self.__health_methods = {}  # typing: Dict

    def __health_method_wrapper(self, method: Callable):
        try:
            return method() or 'ok', st.HTTP_200_OK
        except BaseException as exception:
            code = getattr(exception, 'status', st.HTTP_500_INTERNAL_SERVER_ERROR)
            return str(exception), code

    def __health_types_filter(self, attr_name: Text) -> bool:
        return attr_name.startswith('check_health_')

    def __get_health_methods_iterator(self) -> Iterable[Tuple[Text, Callable]]:
        filter_method = self.__health_types_filter
        for method_name in filter(filter_method, dir(self)):
            method = getattr(self, method_name)
            if callable(method):
                yield method_name.replace('check_health_', '', 1), method

    def get(self) -> Tuple[Dict, int]:
        if not self.__health_methods:
            self.__health_methods = dict(self.__get_health_methods_iterator())
        result, status = {}, st.HTTP_200_OK
        for key, method in self.__health_methods.items():
            method_result, method_status = self.__health_method_wrapper(method)
            result[key] = method_result
            if method_status > status:
                status = method_status
        return result, status


class DefaultBackend(BaseBackend):
    def check_health_db(self):
        """
        Checking if some database server is unavailable.
        """
        for db_name in self.get_django_settings('DATABASES', {}).keys():
            connections[db_name].ensure_connection()

    def check_health_cache(self):
        """
        Checking ig some cache server is unavailable.
        """
        for cache_name in self.get_django_settings('CACHES', {}).keys():
            self.get_django_cache(cache_name).get('test', 0)

    def check_health_rpc(self):
        """
        At now only checking that RPC enabled or not.
        """
        if not self.get_django_settings('RPC_ENABLED'):
            return 'disabled'
        try:
            celery_app = import_class(
                self.get_django_settings('WORKER_OPTIONS')['app'].replace(':', '.')
            )
        except ImportError:  # nocv
            return "disabled"
        else:
            celery_app.pool.connection.connect()



from collections import OrderedDict

from django import template

register = template.Library()


class IniTag(template.Node):
    def __init__(self, config_name):
        self._config_name = config_name

    def _generate_values(self, values):
        return '\n'.join([f'{k} = {v}' for k, v in values.items()])

    def _iter(self, config):
        return '\n'.join([
            f"[{section}]\n{self._generate_values(values)}"
            for section, values in config.items()
        ]) if isinstance(config, (dict, OrderedDict)) else config

    def render(self, context):
        try:
            return self._iter(context.get(self._config_name.token))
        except Exception as err:
            return f'# Invalid config.\n# {str(err)}'

    @classmethod
    def handle_token(cls, parser, token):
        bits = token.split_contents()

        if len(bits) < 2:  # nocv
            raise template.TemplateSyntaxError(
                "'%s' takes at least one argument (dict with empty sections)" % bits[0])

        return cls(parser.compile_filter(bits[1]))


register.tag(name='ini', compile_function=IniTag.handle_token)
